# SPDX-License-Identifier: MIT
"""
Tests using actual Skala weights loaded from HuggingFace.
"""

import numpy as np
import pytest
import torch

# Enable JAX 64-bit mode
import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp

# PyTorch model loading
from skala.functional import load_functional
from skala.functional.model import SkalaFunctional as TorchSkalaFunctional

# JAX model
from skalax.functional.model import SkalaFunctional as JaxSkalaFunctional
from skalax.convert_weights import (
    load_weights_into_model,
    load_weights_and_buffers_into_model,
)


@pytest.fixture(scope="module")
def loaded_torch_model():
    """Load the pretrained Skala model."""
    return load_functional("skala")


@pytest.fixture(scope="module")
def clean_state_dict(loaded_torch_model):
    """Get clean state dict from loaded model."""
    state_dict = loaded_torch_model.state_dict()
    return {
        k.replace("_traced_model.", ""): v
        for k, v in state_dict.items()
    }


@pytest.fixture(scope="module")
def torch_model_from_weights(clean_state_dict):
    """Create fresh PyTorch model with loaded weights."""
    model = TorchSkalaFunctional(
        lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
    )
    model.load_state_dict(clean_state_dict, strict=True)
    return model


@pytest.fixture(scope="module")
def jax_model_with_weights(clean_state_dict):
    """Create JAX model with loaded weights."""
    key = jax.random.PRNGKey(0)
    jax_model = JaxSkalaFunctional(
        lmax=3,
        non_local=True,
        non_local_hidden_nf=16,
        radius_cutoff=5.0,
        key=key,
    )
    return load_weights_into_model(jax_model, clean_state_dict)


class TestWithLoadedWeights:
    """Tests using actual loaded weights."""

    def test_input_model_with_loaded_weights(
        self, torch_model_from_weights, jax_model_with_weights
    ):
        """Test input model with loaded weights."""
        np.random.seed(42)

        # Create test input
        x_np = np.random.randn(10, 7).astype(np.float64)

        # PyTorch forward
        x_torch = torch.from_numpy(x_np).double()
        torch_model_from_weights.double()
        with torch.no_grad():
            y_torch = torch_model_from_weights.input_model(x_torch).numpy()

        # JAX forward
        x_jax = jnp.array(x_np)
        y_jax = np.array(jax_model_with_weights._input_model(x_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_output_model_with_loaded_weights(
        self, torch_model_from_weights, jax_model_with_weights
    ):
        """Test output model with loaded weights."""
        np.random.seed(42)

        # Create test input (256 + 16 features for non_local=True)
        x_np = np.random.randn(10, 272).astype(np.float64)

        # PyTorch forward
        x_torch = torch.from_numpy(x_np).double()
        torch_model_from_weights.double()
        with torch.no_grad():
            y_torch = torch_model_from_weights.output_model(x_torch).numpy()

        # JAX forward
        x_jax = jnp.array(x_np)
        y_jax = np.array(jax_model_with_weights._output_model(x_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_get_exc_no_nonlocal_with_loaded_weights(self, clean_state_dict):
        """Test get_exc without non-local using loaded weights."""
        np.random.seed(42)

        # Create PyTorch model without non-local
        torch_model = TorchSkalaFunctional(lmax=3, non_local=False)
        # Load only the compatible weights
        torch_model.input_model[0].weight.data = clean_state_dict[
            "input_model.0.weight"
        ]
        torch_model.input_model[0].bias.data = clean_state_dict[
            "input_model.0.bias"
        ]
        torch_model.input_model[2].weight.data = clean_state_dict[
            "input_model.2.weight"
        ]
        torch_model.input_model[2].bias.data = clean_state_dict[
            "input_model.2.bias"
        ]

        # Create JAX model
        key = jax.random.PRNGKey(0)
        jax_model = JaxSkalaFunctional(lmax=3, non_local=False, key=key)
        jax_model = load_weights_into_model(
            jax_model, torch_model.state_dict()
        )

        # Create test molecular features
        n_points = 20
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3),
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(2, 3),
        }

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64))
            for k, v in mol_np.items()
        }
        mol_jax = {
            k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()
        }

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


class TestFullModelWithHuggingFaceWeights:
    """Tests for full model with HuggingFace pretrained weights."""

    def test_all_weights_loaded_correctly(self, clean_state_dict):
        """Verify all 24 weights from HuggingFace are loaded correctly."""
        # Create fresh PyTorch model
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
        )
        torch_model.load_state_dict(clean_state_dict, strict=True)
        torch_model.double()

        # Create JAX model
        key = jax.random.PRNGKey(0)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0, key=key
        )
        jax_model = load_weights_and_buffers_into_model(jax_model, torch_model)

        # Verify input model weights
        np.testing.assert_allclose(
            clean_state_dict["input_model.0.weight"].numpy(),
            np.array(jax_model.input_linear1.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["input_model.0.bias"].numpy(),
            np.array(jax_model.input_linear1.bias),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["input_model.2.weight"].numpy(),
            np.array(jax_model.input_linear2.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["input_model.2.bias"].numpy(),
            np.array(jax_model.input_linear2.bias),
            atol=1e-10,
        )

        # Verify non-local pre-down weights
        np.testing.assert_allclose(
            clean_state_dict["non_local_model.pre_down_layer.0.weight"].numpy(),
            np.array(jax_model.non_local_model.pre_down_linear.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["non_local_model.pre_down_layer.0.bias"].numpy(),
            np.array(jax_model.non_local_model.pre_down_linear.bias),
            atol=1e-10,
        )

        # Verify tp_down weights
        for key in ["weight_0_0_0", "weight_0_1_1", "weight_0_2_2", "weight_0_3_3"]:
            np.testing.assert_allclose(
                clean_state_dict[f"non_local_model.tp_down.{key}"].numpy(),
                np.array(jax_model.non_local_model.tp_down.weights[key]),
                atol=1e-10,
            )

        # Verify tp_up weights
        for key in ["weight_0_0_0", "weight_1_1_0", "weight_2_2_0", "weight_3_3_0"]:
            np.testing.assert_allclose(
                clean_state_dict[f"non_local_model.tp_up.{key}"].numpy(),
                np.array(jax_model.non_local_model.tp_up.weights[key]),
                atol=1e-10,
            )

        # Verify non-local post-up weights
        np.testing.assert_allclose(
            clean_state_dict["non_local_model.post_up_layer.0.weight"].numpy(),
            np.array(jax_model.non_local_model.post_up_linear.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["non_local_model.post_up_layer.0.bias"].numpy(),
            np.array(jax_model.non_local_model.post_up_linear.bias),
            atol=1e-10,
        )

        # Verify output model weights
        np.testing.assert_allclose(
            clean_state_dict["output_model.0.weight"].numpy(),
            np.array(jax_model.output_linear1.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.0.bias"].numpy(),
            np.array(jax_model.output_linear1.bias),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.2.weight"].numpy(),
            np.array(jax_model.output_linear2.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.2.bias"].numpy(),
            np.array(jax_model.output_linear2.bias),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.4.weight"].numpy(),
            np.array(jax_model.output_linear3.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.4.bias"].numpy(),
            np.array(jax_model.output_linear3.bias),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.6.weight"].numpy(),
            np.array(jax_model.output_linear4.weight),
            atol=1e-10,
        )
        np.testing.assert_allclose(
            clean_state_dict["output_model.6.bias"].numpy(),
            np.array(jax_model.output_linear4.bias),
            atol=1e-10,
        )

    def test_get_exc_with_pretrained_weights(self, clean_state_dict):
        """Test get_exc with full pretrained model including non-local."""
        np.random.seed(42)

        # Create fresh PyTorch model
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
        )
        torch_model.load_state_dict(clean_state_dict, strict=True)
        torch_model.double()

        # Create JAX model
        key = jax.random.PRNGKey(0)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0, key=key
        )
        jax_model = load_weights_and_buffers_into_model(jax_model, torch_model)

        # Create test molecular features
        n_points = 50
        n_atoms = 5
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3) * 3,
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(n_atoms, 3) * 2,
        }

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64)) for k, v in mol_np.items()
        }
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()}

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_get_exc_density_with_pretrained_weights(self, clean_state_dict):
        """Test get_exc_density with full pretrained model including non-local."""
        np.random.seed(42)

        # Create fresh PyTorch model
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
        )
        torch_model.load_state_dict(clean_state_dict, strict=True)
        torch_model.double()

        # Create JAX model
        key = jax.random.PRNGKey(0)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0, key=key
        )
        jax_model = load_weights_and_buffers_into_model(jax_model, torch_model)

        # Create test molecular features
        n_points = 50
        n_atoms = 5
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3) * 3,
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(n_atoms, 3) * 2,
        }

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64)) for k, v in mol_np.items()
        }
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()}

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc_density(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc_density(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: MIT
"""
Base functions for exchange-correlation functionals in JAX.

This module defines utility functions for implementing exchange-correlation
functionals in Skala JAX.
"""

import jax.numpy as jnp
from jax import Array


# The spin-agnostic version of the lda exchange is
#     E_x(rho) = \int d^3r e_x(rho(r))
#     e_x(rho) = - (3/4) * (3/pi)**(1/3) * rho ** (4/3)
# The spin-polarized version of the lda exchange is
#     E_x(rho_up, rho_down) = 0.5 * \int d^3r e_x(2 rho_up(r)) + e_x(2 rho_down(r))
#                           = 0.5 * 2 ** (4/3) * \int d^3r (e_x(rho_up(r)) + ...
#                           = [ - 2 ** (1/3) * (3/4) * (3/pi)**(1/3) ]
#                              * \int d^3 r rho_up**(4/3) + rho_down**(4/3)
# The prefactor in the squared bracket here is -0.9305257363491001

LDA_PREFACTOR = -0.9305257363491001


def enhancement_density_inner_product(
    enhancement_factor: Array,
    density: Array,
) -> Array:
    """
    Compute the enhancement density as inner product with LDA reference.

    Parameters
    ----------
    enhancement_factor : Array
        Enhancement factor with shape (n, 1).
    density : Array
        Electron density with shape (2, n) for 2 spins and n grid points.

    Returns
    -------
    Array
        Enhanced exchange-correlation density.

    Notes
    -----
    This function computes:
    enhancement_factor * LDA_exchange_density
    where LDA_exchange_density uses the prefactor -0.9305257363491001.
    """
    # Clip density to positive values and compute LDA exchange density
    density_clipped = jnp.clip(density.astype(jnp.float64), 0, None)
    lda = LDA_PREFACTOR * jnp.power(
        density_clipped, 4.0 / 3.0
    ).sum(axis=0).reshape(-1, 1)

    return (enhancement_factor.astype(lda.dtype) * lda).squeeze(1)



# SPDX-License-Identifier: MIT
"""
Tests to verify JAX implementation matches PyTorch implementation.
"""

import numpy as np
import pytest
import torch

# Enable JAX 64-bit mode
import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp

# PyTorch implementations
from skala.functional.layers import ScaledSigmoid as TorchScaledSigmoid
from skala.functional.layers import Squasher as TorchSquasher
from skala.functional.base import (
    enhancement_density_inner_product as torch_enhancement,
)
from skala.functional.base import LDA_PREFACTOR as TORCH_LDA_PREFACTOR
from skala.utils.scatter import scatter_sum as torch_scatter_sum
from skala.functional.model import (
    vect_cdist as torch_vect_cdist,
    exp_radial_func as torch_exp_radial_func,
    polynomial_envelope as torch_polynomial_envelope,
    normalization_envelope as torch_normalization_envelope,
    _prepare_features as torch_prepare_features,
)

# JAX implementations
from skalax.functional.layers import ScaledSigmoid as JaxScaledSigmoid
from skalax.functional.layers import Squasher as JaxSquasher
from skalax.functional.base import (
    enhancement_density_inner_product as jax_enhancement,
)
from skalax.functional.base import LDA_PREFACTOR as JAX_LDA_PREFACTOR
from skalax.utils.scatter import scatter_sum as jax_scatter_sum
from skalax.functional.functions import (
    vect_cdist as jax_vect_cdist,
    exp_radial_func as jax_exp_radial_func,
    polynomial_envelope as jax_polynomial_envelope,
    normalization_envelope as jax_normalization_envelope,
    prepare_features as jax_prepare_features,
)


class TestLayers:
    """Test custom layers match between PyTorch and JAX."""

    def test_lda_prefactor_matches(self):
        """Verify LDA prefactor constant is identical."""
        assert TORCH_LDA_PREFACTOR == JAX_LDA_PREFACTOR

    @pytest.mark.parametrize("scale", [1.0, 2.0, 0.5])
    def test_scaled_sigmoid(self, scale):
        """Test ScaledSigmoid matches between PyTorch and JAX."""
        np.random.seed(42)
        x_np = np.random.randn(100).astype(np.float64)

        # PyTorch
        torch_layer = TorchScaledSigmoid(scale=scale)
        x_torch = torch.from_numpy(x_np)
        y_torch = torch_layer(x_torch).numpy()

        # JAX
        jax_layer = JaxScaledSigmoid(scale=scale)
        x_jax = jnp.array(x_np)
        y_jax = np.array(jax_layer(x_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)

    @pytest.mark.parametrize("eta", [1e-5, 1e-3, 1e-7])
    def test_squasher(self, eta):
        """Test Squasher matches between PyTorch and JAX."""
        np.random.seed(42)
        x_np = np.random.randn(100).astype(np.float64)

        # PyTorch
        torch_layer = TorchSquasher(eta=eta)
        x_torch = torch.from_numpy(x_np)
        y_torch = torch_layer(x_torch).numpy()

        # JAX
        jax_layer = JaxSquasher(eta=eta)
        x_jax = jnp.array(x_np)
        y_jax = np.array(jax_layer(x_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)


class TestBase:
    """Test base utility functions."""

    def test_enhancement_density_inner_product(self):
        """Test enhancement_density_inner_product matches."""
        np.random.seed(42)

        n_points = 50
        enhancement_factor = np.random.randn(n_points, 1).astype(np.float64)
        density = np.abs(np.random.randn(2, n_points).astype(np.float64)) + 0.1

        # PyTorch
        ef_torch = torch.from_numpy(enhancement_factor)
        density_torch = torch.from_numpy(density)
        y_torch = torch_enhancement(ef_torch, density_torch).numpy()

        # JAX
        ef_jax = jnp.array(enhancement_factor)
        density_jax = jnp.array(density)
        y_jax = np.array(jax_enhancement(ef_jax, density_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)

    def test_enhancement_density_with_negative_density(self):
        """Test enhancement handles negative density (should be clipped)."""
        np.random.seed(42)

        n_points = 50
        enhancement_factor = np.random.randn(n_points, 1).astype(np.float64)
        density = np.random.randn(2, n_points).astype(np.float64)

        # PyTorch
        ef_torch = torch.from_numpy(enhancement_factor)
        density_torch = torch.from_numpy(density)
        y_torch = torch_enhancement(ef_torch, density_torch).numpy()

        # JAX
        ef_jax = jnp.array(enhancement_factor)
        density_jax = jnp.array(density)
        y_jax = np.array(jax_enhancement(ef_jax, density_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)


class TestScatter:
    """Test scatter operations."""

    def test_scatter_sum_1d(self):
        """Test scatter_sum for 1D arrays."""
        np.random.seed(42)

        src = np.random.randn(20).astype(np.float64)
        index = np.array([0, 1, 0, 2, 1, 0, 2, 3, 3, 1,
                         0, 1, 2, 3, 0, 1, 2, 3, 0, 1], dtype=np.int64)
        dim_size = 5

        # PyTorch
        src_torch = torch.from_numpy(src)
        index_torch = torch.from_numpy(index)
        y_torch = torch_scatter_sum(
            src_torch, index_torch, dim=0, dim_size=dim_size
        ).numpy()

        # JAX
        src_jax = jnp.array(src)
        index_jax = jnp.array(index)
        y_jax = np.array(jax_scatter_sum(
            src_jax, index_jax, dim=0, dim_size=dim_size
        ))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)

    def test_scatter_sum_2d(self):
        """Test scatter_sum for 2D arrays."""
        np.random.seed(42)

        src = np.random.randn(10, 8).astype(np.float64)
        index = np.array([0, 1, 0, 2, 1, 0, 2, 3, 3, 1], dtype=np.int64)
        dim_size = 5

        # PyTorch
        src_torch = torch.from_numpy(src)
        index_torch = torch.from_numpy(index)
        y_torch = torch_scatter_sum(
            src_torch, index_torch, dim=0, dim_size=dim_size
        ).numpy()

        # JAX
        src_jax = jnp.array(src)
        index_jax = jnp.array(index)
        y_jax = np.array(jax_scatter_sum(
            src_jax, index_jax, dim=0, dim_size=dim_size
        ))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)


class TestFunctions:
    """Test utility functions match between PyTorch and JAX."""

    def test_vect_cdist(self):
        """Test vect_cdist matches."""
        np.random.seed(42)
        c1 = np.random.randn(10, 3).astype(np.float64)
        c2 = np.random.randn(5, 3).astype(np.float64)

        # PyTorch
        c1_torch = torch.from_numpy(c1)
        c2_torch = torch.from_numpy(c2)
        dir_torch, dist_torch = torch_vect_cdist(c1_torch, c2_torch)
        dir_torch = dir_torch.numpy()
        dist_torch = dist_torch.numpy()

        # JAX
        c1_jax = jnp.array(c1)
        c2_jax = jnp.array(c2)
        dir_jax, dist_jax = jax_vect_cdist(c1_jax, c2_jax)
        dir_jax = np.array(dir_jax)
        dist_jax = np.array(dist_jax)

        np.testing.assert_allclose(dir_torch, dir_jax, atol=1e-10)
        np.testing.assert_allclose(dist_torch, dist_jax, atol=1e-10)

    @pytest.mark.parametrize("num_basis", [8, 16, 32])
    def test_exp_radial_func(self, num_basis):
        """Test exp_radial_func matches."""
        np.random.seed(42)
        dist = np.abs(np.random.randn(50).astype(np.float64)) + 0.1

        # PyTorch
        dist_torch = torch.from_numpy(dist)
        y_torch = torch_exp_radial_func(dist_torch, num_basis).numpy()

        # JAX
        dist_jax = jnp.array(dist)
        y_jax = np.array(jax_exp_radial_func(dist_jax, num_basis))

        # Tolerance of 1e-5 as agreed (differences ~1e-7 due to linspace impl)
        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    @pytest.mark.parametrize("cutoff", [3.0, 5.0, 10.0])
    @pytest.mark.parametrize("p", [2, 4, 8])
    def test_polynomial_envelope(self, cutoff, p):
        """Test polynomial_envelope matches."""
        np.random.seed(42)
        r = np.abs(np.random.randn(50).astype(np.float64)) * cutoff

        # PyTorch
        r_torch = torch.from_numpy(r)
        y_torch = torch_polynomial_envelope(r_torch, cutoff, p).numpy()

        # JAX
        r_jax = jnp.array(r)
        y_jax = np.array(jax_polynomial_envelope(r_jax, cutoff, p))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)

    @pytest.mark.parametrize("cutoff", [3.0, 5.0, 10.0])
    def test_normalization_envelope(self, cutoff):
        """Test normalization_envelope matches."""
        np.random.seed(42)
        r = np.abs(np.random.randn(50).astype(np.float64)) * cutoff

        # PyTorch
        r_torch = torch.from_numpy(r)
        y_torch = torch_normalization_envelope(r_torch, cutoff).numpy()

        # JAX
        r_jax = jnp.array(r)
        y_jax = np.array(jax_normalization_envelope(r_jax, cutoff))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-10)

    def test_prepare_features(self):
        """Test prepare_features matches."""
        np.random.seed(42)
        n_points = 50

        mol_np = {
            "density": np.abs(np.random.randn(2, n_points).astype(np.float64)),
            "grad": np.random.randn(2, 3, n_points).astype(np.float64),
            "kin": np.abs(np.random.randn(2, n_points).astype(np.float64)),
        }

        # PyTorch
        mol_torch = {
            k: torch.from_numpy(v) for k, v in mol_np.items()
        }
        feat_ab_torch, feat_ba_torch = torch_prepare_features(mol_torch)
        feat_ab_torch = feat_ab_torch.numpy()
        feat_ba_torch = feat_ba_torch.numpy()

        # JAX
        mol_jax = {
            k: jnp.array(v) for k, v in mol_np.items()
        }
        feat_ab_jax, feat_ba_jax = jax_prepare_features(mol_jax)
        feat_ab_jax = np.array(feat_ab_jax)
        feat_ba_jax = np.array(feat_ba_jax)

        np.testing.assert_allclose(feat_ab_torch, feat_ab_jax, atol=1e-10)
        np.testing.assert_allclose(feat_ba_torch, feat_ba_jax, atol=1e-10)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: MIT
"""
Skala JAX - JAX implementation of the Skala neural XC functional.

This package provides a JAX/Equinox implementation of the Skala model,
numerically equivalent to the PyTorch version for identical weights.

Example
-------
>>> import jax
>>> jax.config.update("jax_enable_x64", True)
>>>
>>> from skalax import SkalaFunctional, load_weights_from_npz, load_config, get_default_weights_dir
>>>
>>> # Load pretrained weights
>>> weights_dir = get_default_weights_dir()
>>> config = load_config(weights_dir)
>>>
>>> key = jax.random.PRNGKey(0)
>>> model = SkalaFunctional(
...     lmax=config["lmax"],
...     non_local=config["non_local"],
...     non_local_hidden_nf=config["non_local_hidden_nf"],
...     radius_cutoff=config["radius_cutoff"],
...     key=key,
... )
>>> model = load_weights_from_npz(model, weights_dir)
"""

__version__ = "1.0.0"

from skalax.functional.model import SkalaFunctional
from skalax.convert_weights import (
    load_weights_from_npz,
    load_config,
    get_default_weights_dir,
)

__all__ = [
    "SkalaFunctional",
    "load_weights_from_npz",
    "load_config",
    "get_default_weights_dir",
]



# SPDX-License-Identifier: MIT
"""
Utility functions for the Skala JAX model.

Contains distance computations, radial basis functions, and envelope functions.
"""

import math

import jax
import jax.numpy as jnp
from jax import Array


# Covalent radius constants (Angstrom to Bohr conversion)
# 0.32 and 2.32 are the smallest and largest covalent radius estimates
# from Pyykko and Atsumi, Chem. Eur. J. 15, 2009, 188-197
ANGSTROM_TO_BOHR = 1.88973
MIN_COV_RAD = 0.32 * ANGSTROM_TO_BOHR
MAX_COV_RAD = 2.32 * ANGSTROM_TO_BOHR


def vect_cdist(c1: Array, c2: Array) -> tuple[Array, Array]:
    """
    Compute pairwise direction vectors and distances between point sets.

    Parameters
    ----------
    c1 : Array
        First set of coordinates, shape (n, 3).
    c2 : Array
        Second set of coordinates, shape (m, 3).

    Returns
    -------
    tuple[Array, Array]
        - direction: Normalized direction vectors, shape (n, m, 3)
        - dist: Distances, shape (n, m)
    """
    direction = c1[:, None] - c2[None, :]
    dist = jnp.sqrt((direction**2 + 1e-20).sum(-1))
    return direction / dist[:, :, None], dist


def exp_radial_func(dist: Array, num_basis: int, dim: int = 3) -> Array:
    """
    Compute exponential radial basis functions.

    This ensures two standard deviations of the Gaussian kernel would reach
    the desired covalent radius value (95% of the Gaussian mass).

    Parameters
    ----------
    dist : Array
        Distances, shape (...).
    num_basis : int
        Number of basis functions.
    dim : int
        Dimensionality for normalization. Default: 3.

    Returns
    -------
    Array
        Radial basis values, shape (..., num_basis).
    """
    min_std = MIN_COV_RAD / 2
    max_std = MAX_COV_RAD / 2
    s = jnp.linspace(min_std, max_std, num_basis)

    temps = 2 * s**2
    x2 = dist[..., None] ** 2
    emb = (
        jnp.exp(-x2 / temps)
        * 2
        / dim
        * x2
        / temps
        / (math.pi * temps) ** (0.5 * dim)
    )

    return emb


def polynomial_envelope(r: Array, cutoff: float, p: int) -> Array:
    """
    Compute polynomial envelope function.

    This smoothly maps the domain r=[0, cutoff] to the range [1, 0]
    using a polynomial function. Every r >= cutoff is mapped to 0.

    From DimeNet (https://arxiv.org/abs/2003.03123).

    Parameters
    ----------
    r : Array
        Distances.
    cutoff : float
        Cutoff distance.
    p : int
        Polynomial order, must be >= 2.

    Returns
    -------
    Array
        Envelope values.
    """
    r = r / cutoff
    r = jnp.clip(r, 0, 1)
    x = r - 1
    x2 = x * x
    poly = p * (p + 1) * x2 - 2 * p * x + 2
    return jax.nn.relu(1 - 0.5 * jnp.power(r, p) * poly)


def normalization_envelope(r: Array, cutoff: float) -> Array:
    """
    Compute normalization envelope function.

    Parameters
    ----------
    r : Array
        Distances.
    cutoff : float
        Cutoff distance.

    Returns
    -------
    Array
        Normalization weights.
    """
    r = r / cutoff
    r = jnp.clip(r, 0, 1)
    return 1 - jnp.where(r < 0.5, 2 * r**2, -2 * r**2 + 4 * r - 1)


def prepare_features(mol: dict[str, Array]) -> tuple[Array, Array]:
    """
    Prepare features for the Skala model from molecular data.

    Parameters
    ----------
    mol : dict
        Dictionary containing:
        - density: shape (2, num_grid_points)
        - grad: shape (2, 3, num_grid_points)
        - kin: shape (2, num_grid_points)

    Returns
    -------
    tuple[Array, Array]
        - features_ab: Features with original spin ordering
        - features_ba: Features with swapped spin ordering
    """
    x = jnp.concatenate(
        [
            mol["density"].T,
            (mol["grad"] ** 2).sum(1).T,
            mol["kin"].T,
            (mol["grad"].sum(0) ** 2).sum(0).reshape(-1, 1),
        ],
        axis=1,
    )
    x = x.astype(jnp.float64)

    features = jnp.log(jnp.abs(x) + 1e-5)

    features_ab = features
    features_ba = features[:, jnp.array([1, 0, 3, 2, 5, 4, 6])]
    return features_ab, features_ba



# SPDX-License-Identifier: MIT
"""
Tests for NonLocalModel and TensorProduct components.
"""

import numpy as np
import pytest
import torch

# Enable JAX 64-bit mode
import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp

# PyTorch model
from skala.functional.model import (
    SkalaFunctional as TorchSkalaFunctional,
    TensorProduct as TorchTensorProduct,
    NonLocalModel as TorchNonLocalModel,
)
from e3nn import o3

# JAX model
from skalax.functional.model import (
    SkalaFunctional as JaxSkalaFunctional,
    TensorProduct as JaxTensorProduct,
    NonLocalModel as JaxNonLocalModel,
)
from skalax.convert_weights import load_weights_and_buffers_into_model
import e3nn_jax as e3nn


class TestTensorProductEquivalence:
    """Test TensorProduct matches between PyTorch and JAX."""

    def test_tensor_product_with_loaded_weights(self):
        """Test TensorProduct with weights loaded from PyTorch."""
        np.random.seed(42)

        # Create PyTorch TensorProduct
        torch_irreps_in1 = o3.Irreps("16x0e")
        torch_irreps_in2 = o3.Irreps.spherical_harmonics(3, p=1)
        torch_irreps_out = o3.Irreps("+".join([f"16x{i}e" for i in range(4)]))

        torch_tp = TorchTensorProduct(
            torch_irreps_in1, torch_irreps_in2, torch_irreps_out
        )
        torch_tp.double()  # Convert to double precision

        # Create JAX TensorProduct
        jax_irreps_in1 = e3nn.Irreps("16x0e")
        jax_irreps_in2 = e3nn.Irreps.spherical_harmonics(3, p=1)
        jax_irreps_out = e3nn.Irreps("+".join([f"16x{i}e" for i in range(4)]))

        key = jax.random.PRNGKey(0)
        jax_tp = JaxTensorProduct(
            jax_irreps_in1, jax_irreps_in2, jax_irreps_out, key=key
        )

        # Load PyTorch weights into JAX model
        # Get weights
        jax_weights = {}
        for name, param in torch_tp.named_parameters():
            jax_weights[name] = jnp.array(param.detach().numpy())

        # Get buffers (w3j)
        jax_w3j = {}
        for name, buf in torch_tp.named_buffers():
            jax_w3j[name] = jnp.array(buf.detach().numpy())

        # Update JAX model
        import equinox as eqx
        jax_tp = eqx.tree_at(lambda tp: tp.weights, jax_tp, jax_weights)
        jax_tp = eqx.tree_at(lambda tp: tp.w3j, jax_tp, jax_w3j)

        # Create test input
        batch_size = 10
        x1_np = np.random.randn(batch_size, 16).astype(np.float64)
        x2_np = np.random.randn(batch_size, 16).astype(np.float64)

        # PyTorch forward
        x1_torch = torch.from_numpy(x1_np)
        x2_torch = torch.from_numpy(x2_np)
        with torch.no_grad():
            y_torch = torch_tp(x1_torch, x2_torch).numpy()

        # JAX forward
        x1_jax = jnp.array(x1_np)
        x2_jax = jnp.array(x2_np)
        y_jax = np.array(jax_tp(x1_jax, x2_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


class TestNonLocalModelEquivalence:
    """Test NonLocalModel matches between PyTorch and JAX."""

    def test_nonlocal_model_with_loaded_weights(self):
        """Test NonLocalModel with weights loaded from PyTorch."""
        np.random.seed(42)

        # Create PyTorch NonLocalModel
        torch_nl = TorchNonLocalModel(
            input_nf=256,
            hidden_nf=16,
            lmax=3,
            radius_cutoff=5.0,
        )
        torch_nl.double()

        # Create JAX NonLocalModel
        key = jax.random.PRNGKey(0)
        jax_nl = JaxNonLocalModel(
            input_nf=256,
            hidden_nf=16,
            lmax=3,
            radius_cutoff=5.0,
            key=key,
        )

        # Load PyTorch weights into JAX model
        import equinox as eqx

        # Pre-down layer
        pre_w = jnp.array(torch_nl.pre_down_layer[0].weight.detach().numpy())
        pre_b = jnp.array(torch_nl.pre_down_layer[0].bias.detach().numpy())
        jax_nl = eqx.tree_at(
            lambda m: (m.pre_down_linear.weight, m.pre_down_linear.bias),
            jax_nl,
            (pre_w, pre_b),
        )

        # Post-up layer
        post_w = jnp.array(torch_nl.post_up_layer[0].weight.detach().numpy())
        post_b = jnp.array(torch_nl.post_up_layer[0].bias.detach().numpy())
        jax_nl = eqx.tree_at(
            lambda m: (m.post_up_linear.weight, m.post_up_linear.bias),
            jax_nl,
            (post_w, post_b),
        )

        # TensorProduct tp_down weights and w3j
        tp_down_weights = {}
        tp_down_w3j = {}
        for name, param in torch_nl.tp_down.named_parameters():
            tp_down_weights[name] = jnp.array(param.detach().numpy())
        for name, buf in torch_nl.tp_down.named_buffers():
            tp_down_w3j[name] = jnp.array(buf.detach().numpy())

        jax_nl = eqx.tree_at(lambda m: m.tp_down.weights, jax_nl, tp_down_weights)
        jax_nl = eqx.tree_at(lambda m: m.tp_down.w3j, jax_nl, tp_down_w3j)

        # TensorProduct tp_up weights and w3j
        tp_up_weights = {}
        tp_up_w3j = {}
        for name, param in torch_nl.tp_up.named_parameters():
            tp_up_weights[name] = jnp.array(param.detach().numpy())
        for name, buf in torch_nl.tp_up.named_buffers():
            tp_up_w3j[name] = jnp.array(buf.detach().numpy())

        jax_nl = eqx.tree_at(lambda m: m.tp_up.weights, jax_nl, tp_up_weights)
        jax_nl = eqx.tree_at(lambda m: m.tp_up.w3j, jax_nl, tp_up_w3j)

        # Create test input
        num_fine = 20
        num_coarse = 3
        h_np = np.random.randn(num_fine, 256).astype(np.float64)
        grid_coords_np = np.random.randn(num_fine, 3).astype(np.float64)
        coarse_coords_np = np.random.randn(num_coarse, 3).astype(np.float64)
        grid_weights_np = np.abs(np.random.randn(num_fine)).astype(np.float64) + 0.1

        # PyTorch forward
        h_torch = torch.from_numpy(h_np)
        grid_coords_torch = torch.from_numpy(grid_coords_np)
        coarse_coords_torch = torch.from_numpy(coarse_coords_np)
        grid_weights_torch = torch.from_numpy(grid_weights_np)

        with torch.no_grad():
            y_torch = torch_nl(
                h_torch, grid_coords_torch, coarse_coords_torch, grid_weights_torch
            ).numpy()

        # JAX forward
        h_jax = jnp.array(h_np)
        grid_coords_jax = jnp.array(grid_coords_np)
        coarse_coords_jax = jnp.array(coarse_coords_np)
        grid_weights_jax = jnp.array(grid_weights_np)

        y_jax = np.array(
            jax_nl(h_jax, grid_coords_jax, coarse_coords_jax, grid_weights_jax)
        )

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


class TestFullModelWithNonLocal:
    """Test full model with non-local component."""

    def test_get_exc_with_nonlocal(self):
        """Test get_exc with non-local model."""
        np.random.seed(42)

        # Create PyTorch model
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
        )

        # Create JAX model
        key = jax.random.PRNGKey(0)
        jax_model = JaxSkalaFunctional(
            lmax=3,
            non_local=True,
            non_local_hidden_nf=16,
            radius_cutoff=5.0,
            key=key,
        )

        # Load weights and buffers
        jax_model = load_weights_and_buffers_into_model(jax_model, torch_model)

        # Create test molecular features
        n_points = 30
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3),
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(3, 3),
        }

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64)) for k, v in mol_np.items()
        }
        mol_jax = {
            k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()
        }

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_get_exc_density_with_nonlocal(self):
        """Test get_exc_density with non-local model."""
        np.random.seed(42)

        # Create PyTorch model
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
        )

        # Create JAX model
        key = jax.random.PRNGKey(0)
        jax_model = JaxSkalaFunctional(
            lmax=3,
            non_local=True,
            non_local_hidden_nf=16,
            radius_cutoff=5.0,
            key=key,
        )

        # Load weights and buffers
        jax_model = load_weights_and_buffers_into_model(jax_model, torch_model)

        # Create test molecular features
        n_points = 30
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3),
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(3, 3),
        }

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64)) for k, v in mol_np.items()
        }
        mol_jax = {
            k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()
        }

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc_density(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc_density(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: MIT
# JAX-based DFT classes for PySCF integration

from pyscf import dft, gto

from skalax.pyscf.numint import JaxSkalaNumInt


class JaxSkalaRKS(dft.rks.RKS):
    """Restricted Kohn-Sham DFT with JAX Skala functional.

    This class extends PySCF's RKS to use the JAX implementation of the
    Skala exchange-correlation functional.

    Example
    -------
    >>> from pyscf import gto
    >>> from skalax.pyscf import JaxSkalaRKS
    >>> from skalax.functional.model import SkalaFunctional
    >>> import jax
    >>>
    >>> mol = gto.M(atom="H 0 0 0; H 0 0 1", basis="def2-svp")
    >>> key = jax.random.PRNGKey(0)
    >>> jax_func = SkalaFunctional(lmax=3, non_local=True, key=key)
    >>> # Load weights...
    >>> ks = JaxSkalaRKS(mol, jax_func)
    >>> energy = ks.kernel()
    """

    def __init__(self, mol: gto.Mole, xc):
        """Initialize restricted Kohn-Sham with JAX Skala functional.

        Parameters
        ----------
        mol : gto.Mole
            PySCF molecule object.
        xc : SkalaFunctional
            JAX Skala functional (with loaded weights).
        """
        super().__init__(mol, xc="custom")
        self._keys.add("with_dftd3")
        self._numint = JaxSkalaNumInt(xc)
        self.with_dftd3 = None  # D3 dispersion not supported yet

    def energy_nuc(self):
        """Return nuclear repulsion energy as float for compatibility."""
        return float(super().energy_nuc())


class JaxSkalaUKS(dft.uks.UKS):
    """Unrestricted Kohn-Sham DFT with JAX Skala functional.

    This class extends PySCF's UKS to use the JAX implementation of the
    Skala exchange-correlation functional for open-shell systems.

    Example
    -------
    >>> from pyscf import gto
    >>> from skalax.pyscf import JaxSkalaUKS
    >>> from skalax.functional.model import SkalaFunctional
    >>> import jax
    >>>
    >>> mol = gto.M(atom="O", basis="def2-svp", spin=2)
    >>> key = jax.random.PRNGKey(0)
    >>> jax_func = SkalaFunctional(lmax=3, non_local=True, key=key)
    >>> # Load weights...
    >>> ks = JaxSkalaUKS(mol, jax_func)
    >>> energy = ks.kernel()
    """

    def __init__(self, mol: gto.Mole, xc):
        """Initialize unrestricted Kohn-Sham with JAX Skala functional.

        Parameters
        ----------
        mol : gto.Mole
            PySCF molecule object.
        xc : SkalaFunctional
            JAX Skala functional (with loaded weights).
        """
        super().__init__(mol, xc="custom")
        self._keys.add("with_dftd3")
        self._numint = JaxSkalaNumInt(xc)
        self.with_dftd3 = None  # D3 dispersion not supported yet

    def energy_nuc(self):
        """Return nuclear repulsion energy as float for compatibility."""
        return float(super().energy_nuc())



# SPDX-License-Identifier: MIT
"""
Scatter operations for JAX arrays.

This module provides scatter operations similar to pytorch_scatter,
specifically scatter_sum for aggregating values at specified indices.
"""

import jax
import jax.numpy as jnp
from jax import Array


def scatter_sum(
    src: Array,
    index: Array,
    dim: int = -1,
    dim_size: int | None = None,
) -> Array:
    """
    Sum all values from the src array at indices specified in the index array.

    Parameters
    ----------
    src : Array
        Source array containing values to scatter.
    index : Array
        Index array specifying where to scatter values.
    dim : int, optional
        Dimension along which to scatter. Default: -1.
    dim_size : int or None, optional
        Size of the output array along the scatter dimension.

    Returns
    -------
    Array
        Array with scattered and summed values.
    """
    # Handle negative dimension
    if dim < 0:
        dim = src.ndim + dim

    # Broadcast index to match src shape
    index = broadcast(index, src, dim)

    # Determine output size
    if dim_size is None:
        if index.size == 0:
            dim_size = 0
        else:
            dim_size = int(index.max()) + 1

    # Create output shape
    out_shape = list(src.shape)
    out_shape[dim] = dim_size

    # Create zeros output
    out = jnp.zeros(out_shape, dtype=src.dtype)

    # Use segment_sum for simple 1D case, otherwise use at[].add
    if src.ndim == 1:
        return jax.ops.segment_sum(src, index, num_segments=dim_size)
    else:
        # Build slice for scatter operation
        return out.at[_build_scatter_indices(index, dim, src.shape)].add(src)


def _build_scatter_indices(index: Array, dim: int, shape: tuple) -> tuple:
    """Build indices for scatter operation."""
    ndim = len(shape)
    indices = []
    for d in range(ndim):
        if d == dim:
            indices.append(index)
        else:
            # Create range for this dimension and broadcast
            idx = jnp.arange(shape[d])
            # Reshape to broadcast correctly
            reshape_shape = [1] * ndim
            reshape_shape[d] = shape[d]
            idx = idx.reshape(reshape_shape)
            # Broadcast to full shape
            idx = jnp.broadcast_to(idx, shape)
            indices.append(idx)
    return tuple(indices)


def broadcast(src: Array, other: Array, dim: int) -> Array:
    """
    Broadcast src array to match the shape of other array.

    Parameters
    ----------
    src : Array
        Source array to broadcast.
    other : Array
        Target array whose shape to match.
    dim : int
        Dimension along which to perform broadcasting.

    Returns
    -------
    Array
        Broadcasted array with shape matching other.
    """
    if dim < 0:
        dim = other.ndim + dim

    if src.ndim == 1:
        # Add dimensions before dim
        for _ in range(dim):
            src = jnp.expand_dims(src, 0)

    # Add dimensions after to match other's ndim
    while src.ndim < other.ndim:
        src = jnp.expand_dims(src, -1)

    # Broadcast to match other's shape
    src = jnp.broadcast_to(src, other.shape)
    return src



# SPDX-License-Identifier: MIT
"""
Tests for JIT compilation of the full SkalaFunctional model (including non-local).

Verifies that:
1. eqx.filter_jit compiles the model without error
2. JIT output matches eager execution
3. JIT __call__ matches the original _forward_eager implementation
4. JIT + jax.grad works end-to-end
5. JIT'd model still matches PyTorch (numerical equivalence preserved)
"""

import numpy as np
import pytest
import torch

import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp
import equinox as eqx

from skala.functional.model import SkalaFunctional as TorchSkalaFunctional

from skalax.functional.model import (
    SkalaFunctional as JaxSkalaFunctional,
    NonLocalModel as JaxNonLocalModel,
)
from skalax.convert_weights import load_weights_and_buffers_into_model


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

@eqx.filter_jit
def jit_get_exc(model, mol):
    return model.get_exc(mol)


@eqx.filter_jit
def jit_get_exc_density(model, mol):
    return model.get_exc_density(mol)


@eqx.filter_jit
def jit_grad_get_exc(model, mol):
    return jax.grad(model.get_exc)(mol)


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------

@pytest.fixture
def random_mol():
    """Generate a random mol dictionary for testing."""
    np.random.seed(42)
    n_points = 20
    n_atoms = 3
    mol_np = {
        "density": np.abs(np.random.randn(2, n_points)) + 0.1,
        "grad": np.random.randn(2, 3, n_points),
        "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
        "grid_coords": np.random.randn(n_points, 3),
        "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
        "coarse_0_atomic_coords": np.random.randn(n_atoms, 3),
    }
    return mol_np


@pytest.fixture
def torch_jax_models():
    """Create PyTorch and JAX models with identical weights."""
    torch_model = TorchSkalaFunctional(
        lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
    )
    torch_model.double()

    key = jax.random.PRNGKey(0)
    jax_model = JaxSkalaFunctional(
        lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0, key=key
    )
    jax_model = load_weights_and_buffers_into_model(jax_model, torch_model)
    return torch_model, jax_model


@pytest.fixture
def nonlocal_model_with_weights():
    """Create a NonLocalModel with PyTorch-transferred weights."""
    from skala.functional.model import NonLocalModel as TorchNonLocalModel

    torch_nl = TorchNonLocalModel(
        input_nf=256, hidden_nf=16, lmax=3, radius_cutoff=5.0
    )
    torch_nl.double()

    key = jax.random.PRNGKey(0)
    jax_nl = JaxNonLocalModel(
        input_nf=256, hidden_nf=16, lmax=3, radius_cutoff=5.0, key=key
    )

    # Transfer weights
    pre_w = jnp.array(torch_nl.pre_down_layer[0].weight.detach().numpy())
    pre_b = jnp.array(torch_nl.pre_down_layer[0].bias.detach().numpy())
    jax_nl = eqx.tree_at(
        lambda m: (m.pre_down_linear.weight, m.pre_down_linear.bias),
        jax_nl, (pre_w, pre_b),
    )

    post_w = jnp.array(torch_nl.post_up_layer[0].weight.detach().numpy())
    post_b = jnp.array(torch_nl.post_up_layer[0].bias.detach().numpy())
    jax_nl = eqx.tree_at(
        lambda m: (m.post_up_linear.weight, m.post_up_linear.bias),
        jax_nl, (post_w, post_b),
    )

    for attr, torch_tp in [("tp_down", torch_nl.tp_down), ("tp_up", torch_nl.tp_up)]:
        weights = {n: jnp.array(p.detach().numpy()) for n, p in torch_tp.named_parameters()}
        w3j = {n: jnp.array(b.detach().numpy()) for n, b in torch_tp.named_buffers()}
        jax_nl = eqx.tree_at(lambda m, a=attr: getattr(m, a).weights, jax_nl, weights)
        jax_nl = eqx.tree_at(lambda m, a=attr: getattr(m, a).w3j, jax_nl, w3j)

    return torch_nl, jax_nl


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------

class TestJITCompilation:
    """Test that eqx.filter_jit compiles the full model without error."""

    def test_jit_get_exc(self, torch_jax_models, random_mol):
        """eqx.filter_jit(model.get_exc) compiles and runs."""
        _, jax_model = torch_jax_models
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        result = jit_get_exc(jax_model, mol_jax)
        assert result.shape == ()
        assert jnp.isfinite(result)

    def test_jit_get_exc_density(self, torch_jax_models, random_mol):
        """eqx.filter_jit(model.get_exc_density) compiles and runs."""
        _, jax_model = torch_jax_models
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        result = jit_get_exc_density(jax_model, mol_jax)
        n_points = random_mol["density"].shape[1]
        assert result.shape == (n_points,)
        assert jnp.all(jnp.isfinite(result))


class TestJITMatchesEager:
    """Test that JIT output matches eager execution."""

    def test_get_exc_jit_vs_eager(self, torch_jax_models, random_mol):
        """JIT'd get_exc matches eager get_exc."""
        _, jax_model = torch_jax_models
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        eager_result = jax_model.get_exc(mol_jax)
        jit_result = jit_get_exc(jax_model, mol_jax)

        np.testing.assert_allclose(
            float(eager_result), float(jit_result), atol=1e-10
        )

    def test_get_exc_density_jit_vs_eager(self, torch_jax_models, random_mol):
        """JIT'd get_exc_density matches eager get_exc_density."""
        _, jax_model = torch_jax_models
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        eager_result = np.array(jax_model.get_exc_density(mol_jax))
        jit_result = np.array(jit_get_exc_density(jax_model, mol_jax))

        np.testing.assert_allclose(eager_result, jit_result, atol=1e-10)


class TestJITMatchesForwardEager:
    """Test that the JIT-compatible __call__ matches _forward_eager."""

    def test_nonlocal_call_vs_forward_eager(self, nonlocal_model_with_weights):
        """NonLocalModel.__call__ matches _forward_eager."""
        np.random.seed(42)
        _, jax_nl = nonlocal_model_with_weights

        num_fine, num_coarse = 20, 3
        h = jnp.array(np.random.randn(num_fine, 256).astype(np.float64))
        grid_coords = jnp.array(np.random.randn(num_fine, 3).astype(np.float64))
        coarse_coords = jnp.array(np.random.randn(num_coarse, 3).astype(np.float64))
        grid_weights = jnp.array(
            (np.abs(np.random.randn(num_fine)) + 0.1).astype(np.float64)
        )

        result_call = np.array(jax_nl(h, grid_coords, coarse_coords, grid_weights))
        result_eager = np.array(
            jax_nl._forward_eager(h, grid_coords, coarse_coords, grid_weights)
        )

        np.testing.assert_allclose(result_call, result_eager, atol=1e-10)

    def test_nonlocal_jit_call_vs_forward_eager(self, nonlocal_model_with_weights):
        """JIT'd NonLocalModel.__call__ matches _forward_eager."""
        np.random.seed(42)
        _, jax_nl = nonlocal_model_with_weights

        num_fine, num_coarse = 20, 3
        h = jnp.array(np.random.randn(num_fine, 256).astype(np.float64))
        grid_coords = jnp.array(np.random.randn(num_fine, 3).astype(np.float64))
        coarse_coords = jnp.array(np.random.randn(num_coarse, 3).astype(np.float64))
        grid_weights = jnp.array(
            (np.abs(np.random.randn(num_fine)) + 0.1).astype(np.float64)
        )

        @eqx.filter_jit
        def jit_nl(model, h, gc, cc, gw):
            return model(h, gc, cc, gw)

        result_jit = np.array(jit_nl(jax_nl, h, grid_coords, coarse_coords, grid_weights))
        result_eager = np.array(
            jax_nl._forward_eager(h, grid_coords, coarse_coords, grid_weights)
        )

        np.testing.assert_allclose(result_jit, result_eager, atol=1e-10)


class TestJITGrad:
    """Test that JIT + jax.grad works."""

    def test_jit_grad_get_exc(self, torch_jax_models, random_mol):
        """eqx.filter_jit(jax.grad(model.get_exc)) produces finite gradients."""
        _, jax_model = torch_jax_models
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        grads = jit_grad_get_exc(jax_model, mol_jax)

        for key in ["density", "grad", "kin", "grid_weights"]:
            assert jnp.all(jnp.isfinite(grads[key])), f"Non-finite gradient for {key}"

    def test_jit_grad_matches_eager_grad(self, torch_jax_models, random_mol):
        """JIT'd gradients match eager gradients."""
        _, jax_model = torch_jax_models
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        eager_grads = jax.grad(jax_model.get_exc)(mol_jax)
        jit_grads = jit_grad_get_exc(jax_model, mol_jax)

        for key in ["density", "grad", "kin", "grid_weights"]:
            np.testing.assert_allclose(
                np.array(eager_grads[key]),
                np.array(jit_grads[key]),
                atol=1e-10,
                err_msg=f"Gradient mismatch for {key}",
            )


class TestJITPreservesPyTorchEquivalence:
    """Test that JIT'd model still matches PyTorch."""

    def test_get_exc_jit_matches_pytorch(self, torch_jax_models, random_mol):
        """JIT'd JAX get_exc matches PyTorch get_exc."""
        torch_model, jax_model = torch_jax_models

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64)) for k, v in random_mol.items()
        }
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        with torch.no_grad():
            y_torch = torch_model.get_exc(mol_torch).numpy()

        y_jax = float(jit_get_exc(jax_model, mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_get_exc_density_jit_matches_pytorch(self, torch_jax_models, random_mol):
        """JIT'd JAX get_exc_density matches PyTorch get_exc_density."""
        torch_model, jax_model = torch_jax_models

        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64)) for k, v in random_mol.items()
        }
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in random_mol.items()}

        with torch.no_grad():
            y_torch = torch_model.get_exc_density(mol_torch).numpy()

        y_jax = np.array(jit_get_exc_density(jax_model, mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_nonlocal_jit_matches_pytorch(self, nonlocal_model_with_weights):
        """JIT'd NonLocalModel matches PyTorch NonLocalModel."""
        np.random.seed(42)
        torch_nl, jax_nl = nonlocal_model_with_weights

        num_fine, num_coarse = 20, 3
        h_np = np.random.randn(num_fine, 256).astype(np.float64)
        grid_coords_np = np.random.randn(num_fine, 3).astype(np.float64)
        coarse_coords_np = np.random.randn(num_coarse, 3).astype(np.float64)
        grid_weights_np = (np.abs(np.random.randn(num_fine)) + 0.1).astype(np.float64)

        with torch.no_grad():
            y_torch = torch_nl(
                torch.from_numpy(h_np),
                torch.from_numpy(grid_coords_np),
                torch.from_numpy(coarse_coords_np),
                torch.from_numpy(grid_weights_np),
            ).numpy()

        @eqx.filter_jit
        def jit_nl(model, h, gc, cc, gw):
            return model(h, gc, cc, gw)

        y_jax = np.array(jit_nl(
            jax_nl,
            jnp.array(h_np),
            jnp.array(grid_coords_np),
            jnp.array(coarse_coords_np),
            jnp.array(grid_weights_np),
        ))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


class TestJITWithPartialCutoff:
    """Test JIT with cases where some pairs are outside the cutoff."""

    def test_sparse_radius_mask(self, torch_jax_models):
        """JIT works when only some pairs are within cutoff."""
        _, jax_model = torch_jax_models

        np.random.seed(123)
        n_points = 15
        # Spread points far apart so some fall outside radius_cutoff=5.0
        mol_jax = {
            "density": jnp.array(np.abs(np.random.randn(2, n_points)) + 0.1),
            "grad": jnp.array(np.random.randn(2, 3, n_points)),
            "kin": jnp.array(np.abs(np.random.randn(2, n_points)) + 0.1),
            "grid_coords": jnp.array(np.random.randn(n_points, 3) * 5.0),
            "grid_weights": jnp.array(np.abs(np.random.randn(n_points)) + 0.1),
            "coarse_0_atomic_coords": jnp.array(np.random.randn(3, 3) * 5.0),
        }

        result = jit_get_exc(jax_model, mol_jax)
        assert jnp.isfinite(result)

        grads = jit_grad_get_exc(jax_model, mol_jax)
        for key in ["density", "grad", "kin", "grid_weights"]:
            assert jnp.all(jnp.isfinite(grads[key]))


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: MIT
# JAX-based NumInt for PySCF integration

# TODO:
# Architecture note: The real bottleneck is that generate_features() is still a PyTorch function. Each SCF iteration goes: PySCF numpy → PyTorch tensor → compute density features → PyTorch tensor → numpy → JAX array → run model → JAX result → numpy → PyTorch tensor → autograd. A pure-JAX feature generator would eliminate the PyTorch↔JAX bridge entirely, but that's a larger refactor.

from collections.abc import Callable
from typing import Any, Protocol

import numpy as np
import jax
import jax.numpy as jnp
import equinox as eqx
import torch
from torch import Tensor
from torch.autograd import Function
from pyscf import dft, gto

from skalax.pyscf.backend import (
    KS,
    Array,
    Grid,
    from_numpy_or_cupy,
    to_numpy,
)
from skalax.pyscf.features import generate_features


# ---------------------------------------------------------------------------
# JIT-compiled wrappers for the JAX model (cached after first SCF iteration)
# ---------------------------------------------------------------------------

@eqx.filter_jit
def _jit_get_exc(model, features_jax):
    """JIT-compiled forward pass: compute E_xc."""
    return model.get_exc(features_jax)


@eqx.filter_jit
def _jit_grad_get_exc(model, features_jax):
    """JIT-compiled backward pass: compute dE_xc/d(features)."""
    return jax.grad(model.get_exc)(features_jax)


class LibXCSpec(Protocol):
    __version__: str | None
    __references__: str | None

    @staticmethod
    def is_hybrid_xc(xc: str) -> bool: ...

    @staticmethod
    def is_nlc(xc: str) -> bool: ...


class JaxExcFunction(Function):
    """Custom autograd function to bridge PyTorch features → JAX model → PyTorch gradients.

    This function enables the chain rule to flow through JAX's autodiff:
    - Forward: Converts PyTorch tensors to JAX arrays, computes E_xc using JAX model
    - Backward: Uses JAX's grad to compute dE_xc/d(features), converts back to PyTorch
    """

    @staticmethod
    def forward(ctx, *args):
        """
        Forward pass: compute E_xc using JAX model.

        args = (*feature_tensors, keys_tuple, jax_func)
        """
        # Last two args are keys and func
        jax_func = args[-1]
        keys = args[-2]
        feature_tensors = args[:-2]

        # Convert PyTorch tensors to JAX arrays via numpy (zero-copy for CPU tensors)
        features_jax = {}
        for key, val in zip(keys, feature_tensors):
            features_jax[key] = jnp.asarray(val.detach().numpy())

        # Compute E_xc using JIT-compiled JAX model
        E_xc = _jit_get_exc(jax_func, features_jax)

        # Save for backward pass
        ctx.features_jax = features_jax
        ctx.jax_func = jax_func
        ctx.keys = keys
        ctx.dtype = feature_tensors[0].dtype
        ctx.device = feature_tensors[0].device

        # .copy() ensures the numpy scalar is writable for PyTorch
        return torch.from_numpy(
            np.asarray(E_xc).copy()
        ).to(dtype=ctx.dtype, device=ctx.device)

    @staticmethod
    def backward(ctx, grad_output):
        """Backward pass: compute gradients using JAX autodiff."""

        # Compute gradients w.r.t. each feature using JIT-compiled JAX
        grads_jax = _jit_grad_get_exc(ctx.jax_func, ctx.features_jax)

        # Convert JAX gradients back to PyTorch tensors via numpy (minimal copy)
        grads_torch = []
        grad_scale = grad_output.item()
        for key in ctx.keys:
            if key in grads_jax and grads_jax[key] is not None:
                grad_np = np.asarray(grads_jax[key])
                if not grad_np.flags.writeable:
                    grad_np = grad_np.copy()
                grad_tensor = torch.from_numpy(grad_np).to(
                    dtype=ctx.dtype, device=ctx.device
                )
                grads_torch.append(grad_tensor * grad_scale)
            else:
                grads_torch.append(None)

        # Return: gradients for features, None for keys and func
        return (*grads_torch, None, None)


class JaxSkalaNumInt:
    """PySCF-compatible NumInt using JAX Skala model.

    This class provides the same interface as `pyscf.dft.numint.NumInt` but uses
    the JAX implementation of the Skala functional for exchange-correlation energy
    and potential computation.

    Example
    -------
    >>> from skalax.pyscf import JaxSkalaKS
    >>> from skalax.functional.model import SkalaFunctional
    >>> import jax
    >>>
    >>> key = jax.random.PRNGKey(0)
    >>> jax_func = SkalaFunctional(lmax=3, non_local=True, key=key)
    >>> # Load weights...
    >>> ks = JaxSkalaKS(mol, xc=jax_func)
    >>> energy = ks.kernel()
    """

    def __init__(self, functional, chunk_size: int | None = None):
        """Initialize the JAX NumInt.

        Parameters
        ----------
        functional : SkalaFunctional
            The JAX Skala functional (with loaded weights).
        chunk_size : int, optional
            Chunk size for grid batching. If None, determined automatically.
        """
        self.func = functional  # JAX SkalaFunctional
        self.chunk_size = chunk_size
        # Features needed by the Skala model
        self.features = {
            "density", "grad", "kin",
            "grid_coords", "grid_weights", "coarse_0_atomic_coords"
        }

    def from_backend(
        self,
        x: Array,
        device: torch.device | None = None,
        transpose: bool = False,
    ) -> Tensor:
        return from_numpy_or_cupy(x, device=device, transpose=transpose)

    def to_backend(self, x: Tensor | list[Tensor]) -> Array | list[Array]:
        if isinstance(x, list):
            return [self.to_backend(y) for y in x]
        return to_numpy(x)

    def __call__(
        self,
        mol: gto.Mole,
        grids: dft.Grids,
        xc_code: str | None,
        dm: Tensor,
        second_order: bool = False,
        max_memory: int = 2000,
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Compute exchange-correlation energy and potential.

        Parameters
        ----------
        mol : gto.Mole
            PySCF molecule object.
        grids : dft.Grids
            Integration grid.
        xc_code : str or None
            XC code (ignored, uses JAX functional).
        dm : Tensor
            Density matrix as PyTorch tensor.
        second_order : bool
            Whether to retain graph for second-order derivatives.
        max_memory : int
            Maximum memory in MB for chunked evaluation.

        Returns
        -------
        N : Tensor
            Number of electrons per spin channel.
        E_xc : Tensor
            Exchange-correlation energy (scalar).
        V_xc : Tensor
            Exchange-correlation potential (same shape as dm).
        """
        dm = dm.requires_grad_()

        # Generate features using PyTorch code (well-tested, handles chunking)
        mol_features = generate_features(
            mol,
            dm,
            grids,
            self.features,
            chunk_size=self.chunk_size,
            max_memory=max_memory,
            gpu=False,  # JAX model runs on CPU for now
        )

        # Prepare for custom autograd function
        keys = tuple(mol_features.keys())
        values = tuple(mol_features[k] for k in keys)

        # Compute E_xc through JAX (with autograd bridge)
        E_xc = JaxExcFunction.apply(*values, keys, self.func)

        # Compute V_xc via PyTorch autograd (chain rule through features)
        (V_xc,) = torch.autograd.grad(
            E_xc,
            dm,
            torch.ones_like(E_xc),
            retain_graph=second_order,
            create_graph=second_order,
        )

        # Compute N (number of electrons)
        rho = mol_features["density"]
        grid_weights = mol_features["grid_weights"]
        N = (rho * grid_weights).sum(dim=-1)

        return N, E_xc, V_xc

    def get_rho(
        self,
        mol: gto.Mole,
        dm: Array,
        grids: Grid,
        max_memory: int = 2000,
        verbose: int = 0,
    ) -> Array:
        """Compute electron density on the grid.

        Parameters
        ----------
        mol : gto.Mole
            PySCF molecule object.
        dm : Array
            Density matrix.
        grids : Grid
            Integration grid.
        max_memory : int
            Maximum memory in MB.
        verbose : int
            Verbosity level (unused).

        Returns
        -------
        rho : Array
            Electron density on the grid.
        """
        mol_features = generate_features(
            mol,
            self.from_backend(dm),
            grids,
            features={"density"},
            chunk_size=self.chunk_size,
            max_memory=max_memory,
            gpu=False,
        )
        return self.to_backend(mol_features["density"].sum(0))

    def nr_rks(
        self,
        mol: gto.Mole,
        grids: Grid,
        xc_code: str | None,
        dm: Array,
        max_memory: int = 2000,
    ) -> tuple[float, float, Array]:
        """Restricted Kohn-Sham method, applicable when both spin-densities are equal.

        Parameters
        ----------
        mol : gto.Mole
            PySCF molecule object.
        grids : Grid
            Integration grid.
        xc_code : str or None
            XC code (ignored).
        dm : Array
            Density matrix (2D array for RKS).
        max_memory : int
            Maximum memory in MB.

        Returns
        -------
        N : float
            Total number of electrons.
        E_xc : float
            Exchange-correlation energy.
        V_xc : Array
            Exchange-correlation potential matrix.
        """
        assert len(dm.shape) == 2
        N, E_xc, V_xc = self(
            mol, grids, xc_code, self.from_backend(dm), max_memory=max_memory
        )
        return N.sum().item(), E_xc.item(), self.to_backend(V_xc)

    def nr_uks(
        self,
        mol: gto.Mole,
        grids: Grid,
        xc_code: str | None,
        dm: Array,
        max_memory: int = 2000,
    ) -> tuple[Array, float, Array]:
        """Unrestricted Kohn-Sham method, spin densities can be different.

        Parameters
        ----------
        mol : gto.Mole
            PySCF molecule object.
        grids : Grid
            Integration grid.
        xc_code : str or None
            XC code (ignored).
        dm : Array
            Density matrices (3D array: [2, nao, nao] for alpha/beta).
        max_memory : int
            Maximum memory in MB.

        Returns
        -------
        N : Array
            Number of electrons per spin channel.
        E_xc : float
            Exchange-correlation energy.
        V_xc : Array
            Exchange-correlation potential matrices.
        """
        assert len(dm.shape) == 3 and dm.shape[0] == 2
        N, E_xc, V_xc = self(
            mol, grids, xc_code, self.from_backend(dm), max_memory=max_memory
        )
        return self.to_backend(N), E_xc.item(), self.to_backend(V_xc)

    def rsh_and_hybrid_coeff(self) -> tuple[float, float, float]:
        """Return range-separation and hybrid coefficients.

        Skala is a pure functional (no exact exchange), so returns zeros.
        """
        return 0, 0, 0

    class libxc:
        """Mock libxc interface for PySCF compatibility."""
        __version__ = None
        __reference__ = None

        @staticmethod
        def is_hybrid_xc(xc: str) -> bool:
            return False

        @staticmethod
        def is_nlc(xc: str) -> bool:
            return False

    def gen_response(
        self,
        mo_coeff: Array | None,
        mo_occ: Array | None,
        *,
        ks: KS,
        **kwargs: Any,
    ) -> Callable[[Array], Array]:
        """Generate the response function for TDDFT/stability analysis.

        Parameters
        ----------
        mo_coeff : Array
            Molecular orbital coefficients.
        mo_occ : Array
            Molecular orbital occupations.
        ks : KS
            Kohn-Sham object.
        **kwargs : dict
            Additional arguments (hermi, singlet, with_j).

        Returns
        -------
        hessian_vector_product : Callable
            Function that computes Hessian-vector products.
        """
        assert mo_coeff is not None
        assert mo_occ is not None

        # Validate kwargs for meta-GGA
        if kwargs is not None:
            if "hermi" in kwargs:
                assert kwargs["hermi"] == 1
            if "singlet" in kwargs:
                assert kwargs["singlet"] is None
            if "with_j" in kwargs:
                assert kwargs["with_j"]

        dm0 = self.from_backend(ks.make_rdm1(mo_coeff, mo_occ))
        # Cache V_xc to save a forward pass in each iteration
        V_xc = self(ks.mol, ks.grids, None, dm0, second_order=True)[2]

        def hessian_vector_product(dm1: Array) -> Array:
            v1 = self.to_backend(
                torch.autograd.grad(
                    V_xc, dm0, self.from_backend(dm1), retain_graph=True
                )[0]
            )
            vj = ks.get_j(ks.mol, dm1, hermi=1)

            if ks.mol.spin == 0:
                v1 += vj
            else:
                v1 += vj[0] + vj[1]

            return v1

        return hessian_vector_product



# SPDX-License-Identifier: MIT
"""
Tests for full model comparison between PyTorch and JAX.
"""

import numpy as np
import pytest
import torch

# Enable JAX 64-bit mode
import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp

# PyTorch model
from skala.functional.model import SkalaFunctional as TorchSkalaFunctional

# JAX model
from skalax.functional.model import SkalaFunctional as JaxSkalaFunctional
from skalax.convert_weights import load_weights_into_model


class TestTensorProduct:
    """Test TensorProduct matches between PyTorch and JAX."""

    def test_tensor_product_output_shape(self):
        """Test TensorProduct output shape matches."""
        import e3nn_jax as e3nn
        from skalax.functional.model import TensorProduct

        key = jax.random.PRNGKey(0)
        irreps_in1 = e3nn.Irreps("16x0e")
        irreps_in2 = e3nn.Irreps.spherical_harmonics(3, p=1)
        irreps_out = e3nn.Irreps("+".join([f"16x{i}e" for i in range(4)]))

        tp = TensorProduct(irreps_in1, irreps_in2, irreps_out, key=key)

        x1 = jnp.ones((5, 16))
        x2 = jnp.ones((5, 16))
        out = tp(x1, x2)

        assert out.shape == (5, 256)


class TestModelInstantiation:
    """Test model can be instantiated."""

    def test_jax_model_instantiation(self):
        """Test JAX model can be created."""
        key = jax.random.PRNGKey(42)
        model = JaxSkalaFunctional(
            lmax=3, non_local=True, radius_cutoff=5.0, key=key
        )
        assert model.non_local is True
        assert model.num_feats == 256

    def test_jax_model_no_nonlocal(self):
        """Test JAX model without non-local."""
        key = jax.random.PRNGKey(42)
        model = JaxSkalaFunctional(
            lmax=3, non_local=False, key=key
        )
        assert model.non_local is False
        assert model.non_local_model is None


class TestWeightLoading:
    """Test weight loading from PyTorch to JAX."""

    def test_load_weights_input_model(self):
        """Test loading weights into input model."""
        # Create PyTorch model
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=True, radius_cutoff=5.0
        )
        torch_state = torch_model.state_dict()

        # Create JAX model
        key = jax.random.PRNGKey(42)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=True, radius_cutoff=5.0, key=key
        )

        # Load weights
        jax_model = load_weights_into_model(jax_model, torch_state)

        # Check input weights match
        torch_w = torch_state["input_model.0.weight"].numpy()
        jax_w = np.array(jax_model.input_linear1.weight)
        np.testing.assert_allclose(torch_w, jax_w, atol=1e-6)


class TestInputModel:
    """Test input model component."""

    def test_input_model_equivalence(self):
        """Test input model produces same output."""
        np.random.seed(42)

        # Create models
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=False, radius_cutoff=5.0
        )

        key = jax.random.PRNGKey(42)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=False, radius_cutoff=5.0, key=key
        )

        # Load PyTorch weights into JAX model
        jax_model = load_weights_into_model(jax_model, torch_model.state_dict())

        # Create test input
        x_np = np.random.randn(10, 7).astype(np.float32)

        # PyTorch forward
        x_torch = torch.from_numpy(x_np)
        with torch.no_grad():
            y_torch = torch_model.input_model(x_torch).numpy()

        # JAX forward
        x_jax = jnp.array(x_np)
        y_jax = np.array(jax_model._input_model(x_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


class TestOutputModel:
    """Test output model component."""

    def test_output_model_equivalence(self):
        """Test output model produces same output."""
        np.random.seed(42)

        # Create models (without non-local for simplicity)
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=False, radius_cutoff=5.0
        )

        key = jax.random.PRNGKey(42)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=False, radius_cutoff=5.0, key=key
        )

        # Load PyTorch weights into JAX model
        jax_model = load_weights_into_model(jax_model, torch_model.state_dict())

        # Create test input (256 features for non-local=False)
        x_np = np.random.randn(10, 256).astype(np.float32)

        # PyTorch forward
        x_torch = torch.from_numpy(x_np)
        with torch.no_grad():
            y_torch = torch_model.output_model(x_torch).numpy()

        # JAX forward
        x_jax = jnp.array(x_np)
        y_jax = np.array(jax_model._output_model(x_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


class TestFullModelNoNonLocal:
    """Test full model without non-local component."""

    def test_get_exc_density_no_nonlocal(self):
        """Test get_exc_density without non-local."""
        np.random.seed(42)

        # Create models
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=False
        )

        key = jax.random.PRNGKey(42)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=False, key=key
        )

        # Load weights
        jax_model = load_weights_into_model(jax_model, torch_model.state_dict())

        # Create test molecular features
        n_points = 20
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3),
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(2, 3),
        }

        # Convert to appropriate types
        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64))
            for k, v in mol_np.items()
        }
        mol_jax = {
            k: jnp.array(v.astype(np.float64))
            for k, v in mol_np.items()
        }

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc_density(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc_density(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)

    def test_get_exc_no_nonlocal(self):
        """Test get_exc without non-local."""
        np.random.seed(42)

        # Create models
        torch_model = TorchSkalaFunctional(
            lmax=3, non_local=False
        )

        key = jax.random.PRNGKey(42)
        jax_model = JaxSkalaFunctional(
            lmax=3, non_local=False, key=key
        )

        # Load weights
        jax_model = load_weights_into_model(jax_model, torch_model.state_dict())

        # Create test molecular features
        n_points = 20
        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3),
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(2, 3),
        }

        # Convert to appropriate types
        mol_torch = {
            k: torch.from_numpy(v.astype(np.float64))
            for k, v in mol_np.items()
        }
        mol_jax = {
            k: jnp.array(v.astype(np.float64))
            for k, v in mol_np.items()
        }

        # Forward pass
        with torch.no_grad():
            y_torch = torch_model.get_exc(mol_torch).numpy()

        y_jax = np.array(jax_model.get_exc(mol_jax))

        np.testing.assert_allclose(y_torch, y_jax, atol=1e-5)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: MIT
"""
Weight conversion utilities for PyTorch to JAX.

Converts PyTorch state_dict to JAX pytree format compatible with Equinox.
Supports loading from PyTorch models or from local .npz files.
"""

import json
from pathlib import Path

import numpy as np
import jax.numpy as jnp
import equinox as eqx


def convert_pytorch_state_dict(state_dict: dict) -> dict:
    """
    Convert PyTorch state_dict to JAX-compatible dictionary.

    Parameters
    ----------
    state_dict : dict
        PyTorch state_dict with tensor values.

    Returns
    -------
    dict
        Dictionary with JAX arrays.
    """
    jax_dict = {}
    for key, tensor in state_dict.items():
        # Convert to numpy then to JAX
        np_array = tensor.detach().cpu().numpy()
        jax_dict[key] = jnp.array(np_array)
    return jax_dict


def extract_buffers(torch_model) -> dict:
    """
    Extract all buffers (including Wigner 3j) from PyTorch model.

    Parameters
    ----------
    torch_model : nn.Module
        PyTorch model.

    Returns
    -------
    dict
        Dictionary with buffer names and JAX arrays.
    """
    buffers = {}
    for name, buf in torch_model.named_buffers():
        buffers[name] = jnp.array(buf.detach().cpu().numpy())
    return buffers


def load_weights_into_model(model, pytorch_state_dict: dict):
    """
    Load PyTorch weights into a JAX/Equinox SkalaFunctional model.

    Parameters
    ----------
    model : SkalaFunctional
        JAX model to load weights into.
    pytorch_state_dict : dict
        PyTorch state_dict.

    Returns
    -------
    SkalaFunctional
        Model with loaded weights.
    """
    jax_dict = convert_pytorch_state_dict(pytorch_state_dict)

    # Helper to update a linear layer
    def update_linear(layer, weight_key, bias_key):
        weight = jax_dict[weight_key]
        bias = jax_dict[bias_key]
        return eqx.tree_at(
            lambda lin: (lin.weight, lin.bias),
            layer,
            (weight, bias),
        )

    # Update input model
    model = eqx.tree_at(
        lambda m: m.input_linear1,
        model,
        update_linear(
            model.input_linear1,
            "input_model.0.weight",
            "input_model.0.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.input_linear2,
        model,
        update_linear(
            model.input_linear2,
            "input_model.2.weight",
            "input_model.2.bias",
        ),
    )

    # Update output model
    model = eqx.tree_at(
        lambda m: m.output_linear1,
        model,
        update_linear(
            model.output_linear1,
            "output_model.0.weight",
            "output_model.0.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.output_linear2,
        model,
        update_linear(
            model.output_linear2,
            "output_model.2.weight",
            "output_model.2.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.output_linear3,
        model,
        update_linear(
            model.output_linear3,
            "output_model.4.weight",
            "output_model.4.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.output_linear4,
        model,
        update_linear(
            model.output_linear4,
            "output_model.6.weight",
            "output_model.6.bias",
        ),
    )

    # Update non-local model if present
    if model.non_local and model.non_local_model is not None:
        # Pre-down layer
        model = eqx.tree_at(
            lambda m: m.non_local_model.pre_down_linear,
            model,
            update_linear(
                model.non_local_model.pre_down_linear,
                "non_local_model.pre_down_layer.0.weight",
                "non_local_model.pre_down_layer.0.bias",
            ),
        )

        # Post-up layer
        model = eqx.tree_at(
            lambda m: m.non_local_model.post_up_linear,
            model,
            update_linear(
                model.non_local_model.post_up_linear,
                "non_local_model.post_up_layer.0.weight",
                "non_local_model.post_up_layer.0.bias",
            ),
        )

        # TensorProduct weights for tp_down
        tp_down_weights = {}
        for key, val in jax_dict.items():
            if key.startswith("non_local_model.tp_down.weight_"):
                weight_key = key.replace("non_local_model.tp_down.", "")
                tp_down_weights[weight_key] = val

        model = eqx.tree_at(
            lambda m: m.non_local_model.tp_down.weights,
            model,
            tp_down_weights,
        )

        # TensorProduct weights for tp_up
        tp_up_weights = {}
        for key, val in jax_dict.items():
            if key.startswith("non_local_model.tp_up.weight_"):
                weight_key = key.replace("non_local_model.tp_up.", "")
                tp_up_weights[weight_key] = val

        model = eqx.tree_at(
            lambda m: m.non_local_model.tp_up.weights,
            model,
            tp_up_weights,
        )

    return model


def load_weights_and_buffers_into_model(model, torch_model):
    """
    Load PyTorch weights AND buffers (Wigner 3j) into JAX model.

    This is the complete loading function that ensures numerical equivalence.

    Parameters
    ----------
    model : SkalaFunctional
        JAX model to load weights into.
    torch_model : nn.Module
        PyTorch model to extract weights and buffers from.

    Returns
    -------
    SkalaFunctional
        Model with loaded weights and buffers.
    """
    # First load the regular weights
    state_dict = {k: v for k, v in torch_model.state_dict().items()}
    model = load_weights_into_model(model, state_dict)

    # Now load the Wigner 3j buffers for non-local model
    if model.non_local and model.non_local_model is not None:
        # Extract buffers from PyTorch model
        buffers = {}
        for name, buf in torch_model.named_buffers():
            buffers[name] = jnp.array(buf.detach().cpu().numpy())

        # Load tp_down w3j buffers
        tp_down_w3j = {}
        for key, val in buffers.items():
            if key.startswith("non_local_model.tp_down.w3j_"):
                w3j_key = key.replace("non_local_model.tp_down.", "")
                tp_down_w3j[w3j_key] = val

        if tp_down_w3j:
            model = eqx.tree_at(
                lambda m: m.non_local_model.tp_down.w3j,
                model,
                tp_down_w3j,
            )

        # Load tp_up w3j buffers
        tp_up_w3j = {}
        for key, val in buffers.items():
            if key.startswith("non_local_model.tp_up.w3j_"):
                w3j_key = key.replace("non_local_model.tp_up.", "")
                tp_up_w3j[w3j_key] = val

        if tp_up_w3j:
            model = eqx.tree_at(
                lambda m: m.non_local_model.tp_up.w3j,
                model,
                tp_up_w3j,
            )

    return model


def load_weights_from_npz(model, weights_dir: Path | str):
    """
    Load weights from local .npz files (no PyTorch or HuggingFace needed).

    Parameters
    ----------
    model : SkalaFunctional
        JAX model to load weights into.
    weights_dir : Path or str
        Directory containing skala_weights.npz, skala_buffers.npz, and config.json

    Returns
    -------
    SkalaFunctional
        Model with loaded weights and buffers.

    Example
    -------
    >>> from skalax.functional.model import SkalaFunctional
    >>> from skalax.convert_weights import load_weights_from_npz
    >>> import jax
    >>>
    >>> key = jax.random.PRNGKey(0)
    >>> model = SkalaFunctional(lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0, key=key)
    >>> model = load_weights_from_npz(model, "skalax/weights")
    """
    weights_dir = Path(weights_dir)

    # Load weights
    weights_file = weights_dir / "skala_weights.npz"
    if not weights_file.exists():
        raise FileNotFoundError(f"Weights file not found: {weights_file}")

    weights_data = np.load(weights_file)
    jax_dict = {k: jnp.array(v) for k, v in weights_data.items()}

    # Helper to update a linear layer
    def update_linear(layer, weight_key, bias_key):
        weight = jax_dict[weight_key]
        bias = jax_dict[bias_key]
        return eqx.tree_at(
            lambda lin: (lin.weight, lin.bias),
            layer,
            (weight, bias),
        )

    # Update input model
    model = eqx.tree_at(
        lambda m: m.input_linear1,
        model,
        update_linear(
            model.input_linear1,
            "input_model.0.weight",
            "input_model.0.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.input_linear2,
        model,
        update_linear(
            model.input_linear2,
            "input_model.2.weight",
            "input_model.2.bias",
        ),
    )

    # Update output model
    model = eqx.tree_at(
        lambda m: m.output_linear1,
        model,
        update_linear(
            model.output_linear1,
            "output_model.0.weight",
            "output_model.0.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.output_linear2,
        model,
        update_linear(
            model.output_linear2,
            "output_model.2.weight",
            "output_model.2.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.output_linear3,
        model,
        update_linear(
            model.output_linear3,
            "output_model.4.weight",
            "output_model.4.bias",
        ),
    )
    model = eqx.tree_at(
        lambda m: m.output_linear4,
        model,
        update_linear(
            model.output_linear4,
            "output_model.6.weight",
            "output_model.6.bias",
        ),
    )

    # Update non-local model if present
    if model.non_local and model.non_local_model is not None:
        # Pre-down layer
        model = eqx.tree_at(
            lambda m: m.non_local_model.pre_down_linear,
            model,
            update_linear(
                model.non_local_model.pre_down_linear,
                "non_local_model.pre_down_layer.0.weight",
                "non_local_model.pre_down_layer.0.bias",
            ),
        )

        # Post-up layer
        model = eqx.tree_at(
            lambda m: m.non_local_model.post_up_linear,
            model,
            update_linear(
                model.non_local_model.post_up_linear,
                "non_local_model.post_up_layer.0.weight",
                "non_local_model.post_up_layer.0.bias",
            ),
        )

        # TensorProduct weights for tp_down
        tp_down_weights = {}
        for key in jax_dict.keys():
            if key.startswith("non_local_model.tp_down.weight_"):
                weight_key = key.replace("non_local_model.tp_down.", "")
                tp_down_weights[weight_key] = jax_dict[key]

        model = eqx.tree_at(
            lambda m: m.non_local_model.tp_down.weights,
            model,
            tp_down_weights,
        )

        # TensorProduct weights for tp_up
        tp_up_weights = {}
        for key in jax_dict.keys():
            if key.startswith("non_local_model.tp_up.weight_"):
                weight_key = key.replace("non_local_model.tp_up.", "")
                tp_up_weights[weight_key] = jax_dict[key]

        model = eqx.tree_at(
            lambda m: m.non_local_model.tp_up.weights,
            model,
            tp_up_weights,
        )

        # Load buffers (Wigner 3j coefficients)
        buffers_file = weights_dir / "skala_buffers.npz"
        if buffers_file.exists():
            buffers_data = np.load(buffers_file)
            buffers = {k: jnp.array(v) for k, v in buffers_data.items()}

            # Load tp_down w3j buffers
            tp_down_w3j = {}
            for key, val in buffers.items():
                if key.startswith("non_local_model.tp_down.w3j_"):
                    w3j_key = key.replace("non_local_model.tp_down.", "")
                    tp_down_w3j[w3j_key] = val

            if tp_down_w3j:
                model = eqx.tree_at(
                    lambda m: m.non_local_model.tp_down.w3j,
                    model,
                    tp_down_w3j,
                )

            # Load tp_up w3j buffers
            tp_up_w3j = {}
            for key, val in buffers.items():
                if key.startswith("non_local_model.tp_up.w3j_"):
                    w3j_key = key.replace("non_local_model.tp_up.", "")
                    tp_up_w3j[w3j_key] = val

            if tp_up_w3j:
                model = eqx.tree_at(
                    lambda m: m.non_local_model.tp_up.w3j,
                    model,
                    tp_up_w3j,
                )

    return model


def load_config(weights_dir: Path | str) -> dict:
    """
    Load model configuration from weights directory.

    Parameters
    ----------
    weights_dir : Path or str
        Directory containing config.json

    Returns
    -------
    dict
        Model configuration.
    """
    weights_dir = Path(weights_dir)
    config_file = weights_dir / "config.json"

    if not config_file.exists():
        # Return default config
        return {
            "lmax": 3,
            "non_local": True,
            "non_local_hidden_nf": 16,
            "radius_cutoff": 5.0,
        }

    with open(config_file) as f:
        return json.load(f)


def get_default_weights_dir():
    """
    Get the path to bundled pretrained weights.

    This function returns the path to the weights directory that is bundled
    with the skala-jax package. These weights are the official Skala pretrained
    weights converted to float64 format.

    Returns
    -------
    Traversable
        Path-like object to the weights directory bundled with the package.

    Example
    -------
    >>> import jax
    >>> jax.config.update("jax_enable_x64", True)
    >>>
    >>> from skalax import SkalaFunctional, load_weights_from_npz, load_config, get_default_weights_dir
    >>>
    >>> weights_dir = get_default_weights_dir()
    >>> config = load_config(weights_dir)
    >>>
    >>> key = jax.random.PRNGKey(0)
    >>> model = SkalaFunctional(
    ...     lmax=config["lmax"],
    ...     non_local=config["non_local"],
    ...     non_local_hidden_nf=config["non_local_hidden_nf"],
    ...     radius_cutoff=config["radius_cutoff"],
    ...     key=key,
    ... )
    >>> model = load_weights_from_npz(model, weights_dir)
    """
    from importlib.resources import files

    return files("skalax") / "weights"



# SPDX-License-Identifier: MIT
"""
Tests for PySCF integration with JAX Skala model.

These tests verify that the JAX model produces consistent results
when used with PySCF for DFT calculations.
"""

import numpy as np
import pytest
import torch

# Enable JAX 64-bit mode
import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp
from pyscf import gto, dft

# PyTorch model loading
from skala.functional import load_functional
from skala.functional.model import SkalaFunctional as TorchSkalaFunctional

# JAX model
from skalax.functional.model import SkalaFunctional as JaxSkalaFunctional
from skalax.convert_weights import load_weights_and_buffers_into_model
from skalax.pyscf import JaxSkalaKS
from skalax.pyscf.numint import JaxExcFunction
from skalax.pyscf.features import generate_features
from skalax.pyscf.backend import from_numpy_or_cupy
from skalax.convert_weights import load_weights_from_npz, load_config, get_default_weights_dir


@pytest.fixture(scope="module")
def torch_model():
    """Load PyTorch model with HuggingFace weights."""
    func_torch = load_functional("skala")
    state_dict = {
        k.replace("_traced_model.", ""): v
        for k, v in func_torch.state_dict().items()
    }
    model = TorchSkalaFunctional(
        lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0
    )
    model.load_state_dict(state_dict, strict=True)
    model.double()
    return model


@pytest.fixture(scope="module")
def jax_model(torch_model):
    """Create JAX model with loaded weights."""
    key = jax.random.PRNGKey(0)
    model = JaxSkalaFunctional(
        lmax=3, non_local=True, non_local_hidden_nf=16, radius_cutoff=5.0, key=key
    )
    return load_weights_and_buffers_into_model(model, torch_model)


@pytest.fixture(scope="module")
def water_molecule():
    """Create water molecule for testing."""
    return gto.M(
        atom="""
        O 0.000000 0.000000 0.000000
        H 0.757000 0.586000 0.000000
        H -0.757000 0.586000 0.000000
        """,
        basis="sto-3g",
        verbose=0
    )


class TestWeightDtypes:
    """Tests for weight dtype consistency."""

    def test_huggingface_weights_are_float32(self):
        """Verify HuggingFace weights are stored as float32."""
        func_torch = load_functional("skala")
        state_dict = {
            k.replace("_traced_model.", ""): v
            for k, v in func_torch.state_dict().items()
        }
        for k, v in state_dict.items():
            assert v.dtype == torch.float32, f"{k} should be float32"

    def test_torch_model_double_converts_to_float64(self, torch_model):
        """Verify .double() converts PyTorch model to float64."""
        for name, param in torch_model.named_parameters():
            assert param.dtype == torch.float64, f"{name} should be float64"

    def test_jax_model_weights_are_float64(self, jax_model):
        """Verify JAX model weights are float64."""
        assert jax_model.input_linear1.weight.dtype == jnp.float64
        assert jax_model.input_linear1.bias.dtype == jnp.float64
        assert jax_model.output_linear1.weight.dtype == jnp.float64


class TestDirectExcComparison:
    """Tests for direct E_xc comparison without PySCF."""

    def test_exc_matches_with_random_input(self, torch_model, jax_model):
        """Test E_xc matches between PyTorch and JAX with random input."""
        np.random.seed(42)
        n_points = 50
        n_atoms = 3

        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3) * 2,
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(n_atoms, 3) * 2,
        }

        mol_torch = {k: torch.from_numpy(v.astype(np.float64)) for k, v in mol_np.items()}
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()}

        with torch.no_grad():
            exc_torch = torch_model.get_exc(mol_torch).numpy()

        exc_jax = np.array(jax_model.get_exc(mol_jax))

        # Should match within ~1e-7 (floating-point ordering differences)
        np.testing.assert_allclose(exc_torch, exc_jax, atol=1e-6)


class TestPySCFFeatureGeneration:
    """Tests for PySCF feature generation."""

    def test_feature_dtypes_are_float64(self, water_molecule):
        """Verify generated features are float64."""
        grids = dft.Grids(water_molecule)
        grids.level = 3
        grids.build()

        dm = dft.RKS(water_molecule).get_init_guess()
        dm_tensor = from_numpy_or_cupy(dm, dtype=torch.float64)

        features = {"density", "grad", "kin", "grid_coords", "grid_weights", "coarse_0_atomic_coords"}
        mol_features = generate_features(
            water_molecule, dm_tensor, grids, features, gpu=False
        )

        for k, v in mol_features.items():
            assert v.dtype == torch.float64, f"{k} should be float64"

    def test_exc_from_pyscf_features_matches(self, torch_model, jax_model, water_molecule):
        """Test E_xc matches when using PySCF-generated features."""
        grids = dft.Grids(water_molecule)
        grids.level = 3
        grids.build()

        dm = dft.RKS(water_molecule).get_init_guess()
        dm_tensor = from_numpy_or_cupy(dm, dtype=torch.float64)

        features = {"density", "grad", "kin", "grid_coords", "grid_weights", "coarse_0_atomic_coords"}
        mol_features = generate_features(
            water_molecule, dm_tensor, grids, features, gpu=False
        )

        # PyTorch
        with torch.no_grad():
            exc_torch = torch_model.get_exc({k: v for k, v in mol_features.items()}).item()

        # JAX
        mol_features_jax = {k: jnp.array(v.detach().numpy()) for k, v in mol_features.items()}
        exc_jax = float(jax_model.get_exc(mol_features_jax))

        # Should match very closely (~1e-9)
        np.testing.assert_allclose(exc_torch, exc_jax, atol=1e-8)


class TestGradientComputation:
    """Tests for gradient computation consistency."""

    def test_gradients_match_between_torch_and_jax(self, torch_model, jax_model):
        """Test gradients dE/d(features) match between PyTorch and JAX."""
        np.random.seed(42)
        n_points = 100
        n_atoms = 3

        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3) * 2,
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(n_atoms, 3) * 2,
        }

        # PyTorch gradients
        mol_torch = {k: torch.from_numpy(v.astype(np.float64)).requires_grad_(True) for k, v in mol_np.items()}
        exc_torch = torch_model.get_exc(mol_torch)
        grads_torch = torch.autograd.grad(exc_torch, list(mol_torch.values()), torch.ones_like(exc_torch))
        grads_torch_dict = {k: g.detach().numpy() for k, g in zip(mol_torch.keys(), grads_torch)}

        # JAX gradients
        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()}
        grads_jax = jax.grad(lambda m: jax_model.get_exc(m))(mol_jax)

        # Compare gradients (should match within ~1e-6)
        for key in ["density", "grad", "kin", "grid_weights"]:
            g_torch = grads_torch_dict[key]
            g_jax = np.array(grads_jax[key])
            rel_diff = np.linalg.norm(g_torch - g_jax) / (np.linalg.norm(g_torch) + 1e-10)
            assert rel_diff < 1e-5, f"Gradient for {key} differs: rel_diff={rel_diff}"

    def test_numerical_gradient_matches_analytical(self, jax_model):
        """Test JAX analytical gradient matches numerical gradient."""
        np.random.seed(42)
        n_points = 50
        n_atoms = 3

        mol_np = {
            "density": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grad": np.random.randn(2, 3, n_points),
            "kin": np.abs(np.random.randn(2, n_points)) + 0.1,
            "grid_coords": np.random.randn(n_points, 3) * 2,
            "grid_weights": np.abs(np.random.randn(n_points)) + 0.1,
            "coarse_0_atomic_coords": np.random.randn(n_atoms, 3) * 2,
        }

        mol_jax = {k: jnp.array(v.astype(np.float64)) for k, v in mol_np.items()}

        # Analytical gradient
        grads_jax = jax.grad(lambda m: jax_model.get_exc(m))(mol_jax)
        analytical_grad = float(grads_jax["density"][0, 0])

        # Numerical gradient
        eps = 1e-6
        mol_plus = {k: v.copy() for k, v in mol_jax.items()}
        mol_minus = {k: v.copy() for k, v in mol_jax.items()}
        mol_plus["density"] = mol_plus["density"].at[0, 0].add(eps)
        mol_minus["density"] = mol_minus["density"].at[0, 0].add(-eps)

        exc_plus = float(jax_model.get_exc(mol_plus))
        exc_minus = float(jax_model.get_exc(mol_minus))
        numerical_grad = (exc_plus - exc_minus) / (2 * eps)

        np.testing.assert_allclose(analytical_grad, numerical_grad, atol=1e-6)


class TestJaxExcFunction:
    """Tests for the JaxExcFunction autograd bridge."""

    def test_jax_exc_function_forward(self, jax_model, water_molecule):
        """Test JaxExcFunction forward pass."""
        grids = dft.Grids(water_molecule)
        grids.level = 3
        grids.build()

        dm = dft.RKS(water_molecule).get_init_guess()
        dm_tensor = from_numpy_or_cupy(dm, dtype=torch.float64)

        features = {"density", "grad", "kin", "grid_coords", "grid_weights", "coarse_0_atomic_coords"}
        mol_features = generate_features(
            water_molecule, dm_tensor, grids, features, gpu=False
        )

        keys = tuple(mol_features.keys())
        values = tuple(mol_features[k] for k in keys)

        exc = JaxExcFunction.apply(*values, keys, jax_model)
        assert exc.dtype == torch.float64
        assert exc.shape == ()
        assert exc.item() < 0  # E_xc should be negative

    def test_jax_exc_function_backward(self, jax_model, water_molecule):
        """Test JaxExcFunction backward pass produces valid gradients."""
        grids = dft.Grids(water_molecule)
        grids.level = 3
        grids.build()

        dm = dft.RKS(water_molecule).get_init_guess()
        dm_tensor = from_numpy_or_cupy(dm, dtype=torch.float64).requires_grad_()

        features = {"density", "grad", "kin", "grid_coords", "grid_weights", "coarse_0_atomic_coords"}
        mol_features = generate_features(
            water_molecule, dm_tensor, grids, features, gpu=False
        )

        keys = tuple(mol_features.keys())
        values = tuple(mol_features[k] for k in keys)

        exc = JaxExcFunction.apply(*values, keys, jax_model)
        vxc = torch.autograd.grad(exc, dm_tensor, torch.ones_like(exc))[0]

        assert vxc.shape == dm_tensor.shape
        assert vxc.dtype == torch.float64
        assert torch.isfinite(vxc).all()


class TestSCFConvergence:
    """Tests for SCF convergence with JAX model."""

    def test_scf_converges(self, jax_model, water_molecule):
        """Test that SCF converges with JAX model."""
        ks = JaxSkalaKS(water_molecule, xc=jax_model)
        ks.verbose = 0
        ks.conv_tol = 1e-9
        energy = ks.kernel()

        assert ks.converged
        assert energy < 0  # Energy should be negative
        assert energy > -100  # Sanity check

    def test_vxc_matches_at_same_dm(self, torch_model, jax_model, water_molecule):
        """Test V_xc matches between PyTorch and JAX at the same density matrix."""
        from skala.pyscf import SkalaKS

        # Run PyTorch to get converged DM
        ks_torch = SkalaKS(water_molecule, xc="skala")
        ks_torch.verbose = 0
        ks_torch.kernel()
        dm_torch = ks_torch.make_rdm1()

        # Setup JAX KS
        ks_jax = JaxSkalaKS(water_molecule, xc=jax_model)
        ks_jax.verbose = 0
        ks_jax.grids.build()

        # Compare E_xc at the same DM
        n_torch, exc_torch, vxc_torch = ks_torch._numint.nr_rks(
            water_molecule, ks_torch.grids, None, dm_torch
        )
        n_jax, exc_jax, vxc_jax = ks_jax._numint.nr_rks(
            water_molecule, ks_jax.grids, None, dm_torch
        )

        # E_xc should match very closely
        np.testing.assert_allclose(exc_torch, exc_jax, atol=1e-8)

        # V_xc should match closely
        np.testing.assert_allclose(vxc_torch, vxc_jax, atol=1e-7)

    def test_energy_difference_is_small(self, jax_model, water_molecule):
        """Test that energy difference between PyTorch and JAX is small."""
        from skala.pyscf import SkalaKS

        # PyTorch
        ks_torch = SkalaKS(water_molecule, xc="skala")
        ks_torch.verbose = 0
        energy_torch = ks_torch.kernel()

        # JAX
        ks_jax = JaxSkalaKS(water_molecule, xc=jax_model)
        ks_jax.verbose = 0
        energy_jax = ks_jax.kernel()

        # Difference should be < 1 mHa
        diff_ha = abs(energy_torch - energy_jax)
        diff_kcal = diff_ha * 627.5

        assert diff_ha < 0.001, (
            f"Energy difference {diff_ha:.6f} Ha ({diff_kcal:.4f} kcal/mol) "
            "is too large"
        )


class TestLocalWeightLoading:
    """Tests for loading weights from local .npz files."""

    def test_load_config(self):
        """Test loading config from local files."""
        weights_dir = get_default_weights_dir()
        config = load_config(weights_dir)

        assert config["lmax"] == 3
        assert config["non_local"] is True
        assert config["non_local_hidden_nf"] == 16
        assert config["radius_cutoff"] == 5.0

    def test_load_weights_from_npz(self):
        """Test loading weights from local .npz files."""
        weights_dir = get_default_weights_dir()
        config = load_config(weights_dir)

        key = jax.random.PRNGKey(0)
        model = JaxSkalaFunctional(
            lmax=config["lmax"],
            non_local=config["non_local"],
            non_local_hidden_nf=config["non_local_hidden_nf"],
            radius_cutoff=config["radius_cutoff"],
            key=key,
        )
        model = load_weights_from_npz(model, weights_dir)

        # Verify model works
        np.random.seed(42)
        n_points = 20
        mol = {
            "density": jnp.array(np.abs(np.random.randn(2, n_points)) + 0.1),
            "grad": jnp.array(np.random.randn(2, 3, n_points)),
            "kin": jnp.array(np.abs(np.random.randn(2, n_points)) + 0.1),
            "grid_coords": jnp.array(np.random.randn(n_points, 3)),
            "grid_weights": jnp.array(np.abs(np.random.randn(n_points)) + 0.1),
            "coarse_0_atomic_coords": jnp.array(np.random.randn(3, 3)),
        }

        exc = model.get_exc(mol)
        assert exc < 0  # E_xc should be negative
        assert jnp.isfinite(exc)

    def test_local_weights_match_huggingface(self, torch_model, jax_model):
        """Test that local weights produce same results as HuggingFace."""
        weights_dir = get_default_weights_dir()
        config = load_config(weights_dir)

        key = jax.random.PRNGKey(0)
        local_model = JaxSkalaFunctional(
            lmax=config["lmax"],
            non_local=config["non_local"],
            non_local_hidden_nf=config["non_local_hidden_nf"],
            radius_cutoff=config["radius_cutoff"],
            key=key,
        )
        local_model = load_weights_from_npz(local_model, weights_dir)

        # Compare with HuggingFace-loaded model
        np.random.seed(42)
        n_points = 50
        mol = {
            "density": jnp.array(np.abs(np.random.randn(2, n_points)) + 0.1),
            "grad": jnp.array(np.random.randn(2, 3, n_points)),
            "kin": jnp.array(np.abs(np.random.randn(2, n_points)) + 0.1),
            "grid_coords": jnp.array(np.random.randn(n_points, 3) * 2),
            "grid_weights": jnp.array(np.abs(np.random.randn(n_points)) + 0.1),
            "coarse_0_atomic_coords": jnp.array(np.random.randn(3, 3) * 2),
        }

        exc_local = float(local_model.get_exc(mol))
        exc_hf = float(jax_model.get_exc(mol))

        # Should be identical (same weights, just loaded differently)
        np.testing.assert_allclose(exc_local, exc_hf, atol=1e-10)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: MIT

"""
PySCF integration for JAX Skala functional.

This module provides seamless integration between the JAX implementation of
Skala exchange-correlation functionals and the PySCF quantum chemistry package,
enabling DFT calculations with neural network-based functionals.
"""

try:
    import pyscf  # noqa: F401
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "PySCF is not installed. Please install it with `pip install pyscf` or `conda install pyscf`."
    ) from e

from pyscf import gto

from skalax.pyscf.dft import JaxSkalaRKS, JaxSkalaUKS


def JaxSkalaKS(mol: gto.Mole, xc):
    """Create a Kohn-Sham calculator using the JAX Skala functional.

    Automatically selects between restricted (RKS) and unrestricted (UKS)
    Kohn-Sham based on the molecule's spin state.

    Parameters
    ----------
    mol : gto.Mole
        The PySCF molecule object.
    xc : SkalaFunctional
        The JAX Skala functional (with loaded weights).

    Returns
    -------
    JaxSkalaRKS or JaxSkalaUKS
        The Kohn-Sham calculator object.

    Example
    -------
    >>> from pyscf import gto
    >>> from skalax.pyscf import JaxSkalaKS
    >>> from skalax.functional.model import SkalaFunctional
    >>> from skalax.convert_weights import load_weights_and_buffers_into_model
    >>> import jax
    >>>
    >>> # Setup molecule
    >>> mol = gto.M(atom="H 0 0 0; H 0 0 1", basis="def2-svp", verbose=0)
    >>>
    >>> # Create and load JAX model
    >>> key = jax.random.PRNGKey(0)
    >>> jax_func = SkalaFunctional(lmax=3, non_local=True, key=key)
    >>> # ... load weights from PyTorch model ...
    >>>
    >>> # Run calculation
    >>> ks = JaxSkalaKS(mol, xc=jax_func)
    >>> energy = ks.kernel()
    """
    if mol.spin == 0:
        return JaxSkalaRKS(mol, xc)
    else:
        return JaxSkalaUKS(mol, xc)


__all__ = ["JaxSkalaKS", "JaxSkalaRKS", "JaxSkalaUKS"]



# SPDX-License-Identifier: MIT
# Copyright (c) Microsoft Corporation.
# Copied from src/skala/pyscf/features.py

"""
Methods for generating and manipulating density features.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import copy

import numpy as np
import torch
from pyscf import dft, gto
from torch import Tensor, nn
from torch.autograd import Function
from torch.autograd.function import FunctionCtx

from skalax.pyscf.backend import (
    Grid,
    check_gpu_imports_were_successful,
    dft_gpu,
    from_numpy_or_cupy,
)

DEFAULT_FEATURES = ["density", "kin", "grad", "grid_coords", "grid_weights"]
DEFAULT_FEATURES_SET = set(DEFAULT_FEATURES)


def maybe_expand_and_divide(
    feature: torch.Tensor, expand: bool, divisor: float
) -> torch.Tensor:
    """
    Expand feature along spin channels and divide its value by divisor if expand is True.
    """
    if expand:
        return torch.stack([feature / divisor, feature / divisor], dim=0)
    else:
        return feature


def generate_features(
    mol: gto.Mole,
    dm: Tensor,
    grids: Grid,
    features: set[str] | None = None,
    chunk_size: int | None = None,
    max_memory: int = 2000,
    gpu: bool = False,
) -> dict[str, Tensor]:
    """Generate density features for a given molecule. The density features are stored in a dictionary
    with the keys matching the requested features.

    Parameters
    ----------
    mol: gto.Mole
      the molecule
    dm: Tensor
      the density matrix
    grids: Grid
      the grid
    features: set[str] | None
      the requested features
    chunk_size: int | None
        a manually specified chunk size for processing the grids, if None the chunk size is determined automatically
    max_memory: int
      the maximum memory to use for calculating the features
    gpu: bool
        whether to use the GPU(4pyscf) for calculations

    Returns
    -------
    dict[str, Tensor]
        A dictionary containing the requested features. The keys are the feature names,
        and the values are the corresponding tensors.
    """
    features = features or DEFAULT_FEATURES_SET

    # if dm is a 3D tensor, then we have a spin-polarized system
    with_spin = True if len(dm.shape) == 3 else False

    if gpu and dm.device.type != "cuda":
        raise ValueError("Density matrix must be on the GPU when gpu=True.")

    mol_features = {}

    if "grid_coords" in features:
        mol_features["grid_coords"] = from_numpy_or_cupy(
            grids.coords, device=dm.device, dtype=dm.dtype
        )

    if "grid_weights" in features:
        mol_features["grid_weights"] = from_numpy_or_cupy(
            grids.weights, device=dm.device, dtype=dm.dtype
        )

    if "coarse_0_atomic_coords" in features:
        mol_features["coarse_0_atomic_coords"] = from_numpy_or_cupy(
            mol.atom_coords(), device=dm.device, dtype=dm.dtype
        )

    with_mgga_feature = (
        "density" in features
        or "grad" in features
        or "kin" in features
        or "lapl" in features
    )
    if with_mgga_feature:
        mgga_features = auto_chunk(
            dm,
            mol,
            grids,
            MGGAFeatureFunction(
                with_density="density" in features,
                with_grad="grad" in features,
                with_kin="kin" in features,
                with_lapl="lapl" in features,
            ),
            block_size=chunk_size,
            max_memory=max_memory,
            fix_block_size=chunk_size is None,
            gpu=gpu,
        )

        for feature in mgga_features:
            mol_features[feature] = maybe_expand_and_divide(
                mgga_features[feature], not with_spin, 2
            )

    return mol_features


def is_density_feature(feature: str) -> bool:
    return feature in {"density", "grad", "kin"}


def partial_feature_function_over_aos(
    feature_function: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    ao: torch.Tensor,
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Returns a function that computes the feature function with the given ao,
    but not the dm already passed to the function.

    Purpose is to allow for chaining of derivatives.
    """

    def partial_feature_function(dm: torch.Tensor) -> torch.Tensor:
        return feature_function(dm, ao)

    return partial_feature_function


def partial_jvp_function_over_tangents(
    func: Callable[[torch.Tensor], torch.Tensor],
    tangents: torch.Tensor,
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Returns a function that computes the jvp of the given function with tangents,
    but not primals already passed to the function.

    Purpose is to allow for chaining of derivatives over primals."""

    def reduced_jvp(primals: torch.Tensor) -> torch.Tensor:
        _, tangent = torch.func.jvp(func, (primals,), (tangents,))
        return tangent

    return reduced_jvp


def partial_vjp_function_over_tangents(
    func: Callable[[torch.Tensor], torch.Tensor],
    tangents: torch.Tensor,
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Returns a function that computes the vjp of the given function with tangents,
    but not primals already passed to the function.

    Purpose is to allow for chaining of derivatives over primals."""

    def reduced_vjp(primals: torch.Tensor) -> torch.Tensor:
        return torch.func.vjp(func, primals)[1](tangents)[0]

    return reduced_vjp


class FeatureFunction(nn.Module, ABC):  # type: ignore[misc]
    deriv: int
    nfeats: int
    only_linear_feats: bool

    @abstractmethod
    def forward(self, dm: torch.Tensor, ao: torch.Tensor) -> torch.Tensor: ...

    @abstractmethod
    def to_dict(self, features: torch.Tensor) -> dict[str, torch.Tensor]: ...


class MGGAFeatureFunction(FeatureFunction):
    with_density: bool
    with_grad: bool
    with_kin: bool
    with_lapl: bool
    with_ked_var: bool
    with_ked_det: bool

    def __init__(
        self,
        with_density: bool = True,
        with_grad: bool = True,
        with_kin: bool = True,
        with_lapl: bool = False,
        with_ked_var: bool = False,
        with_ked_det: bool = False,
    ):
        super().__init__()

        self.with_density = with_density
        self.with_grad = with_grad
        self.with_kin = with_kin
        self.with_lapl = with_lapl
        self.with_ked_var = with_ked_var
        self.with_ked_det = with_ked_det

        self.deriv = 0
        if with_grad or with_kin or with_ked_var or with_ked_det:
            self.deriv = 1
        if with_lapl:
            self.deriv = 2

        self.nfeats = (
            with_density
            + with_grad * 3
            + with_kin
            + with_lapl
            + with_ked_var
            + with_ked_det
        )

        if self.nfeats == 0:
            raise ValueError("At least one feature must be selected.")

        self.only_linear_feats = not (with_ked_var or with_ked_det)

    def to_dict(self, features: torch.Tensor) -> dict[str, torch.Tensor]:
        """Convert the features to a dictionary with the keys being the feature names."""
        feature_index = 0
        feature_dict: dict[str, torch.Tensor] = {}
        if self.with_density:
            feature_dict["density"] = features[..., feature_index, :]
            feature_index += 1
        if self.with_grad:
            feature_dict["grad"] = features[..., feature_index : feature_index + 3, :]
            feature_index += 3
        if self.with_kin:
            feature_dict["kin"] = features[..., feature_index, :]
            feature_index += 1
        if self.with_lapl:
            feature_dict["lapl"] = features[..., feature_index, :]
            feature_index += 1
        if self.with_ked_var:
            feature_dict["ked_var"] = features[..., feature_index, :]
            feature_index += 1
        if self.with_ked_det:
            feature_dict["ked_det"] = features[..., feature_index, :]
            feature_index += 1
        return feature_dict

    def forward(self, dm: torch.Tensor, ao: torch.Tensor) -> torch.Tensor:
        with_Q: bool = self.with_ked_var or self.with_ked_det

        # Flatten all but the last two dimensions
        # then restore the original shape at the end
        dm_view = dm.view(-1, dm.shape[-2], dm.shape[-1])
        # Explicit symmetrization for autodiff
        dm_view = 0.5 * (dm_view + dm_view.transpose(-1, -2))

        features = torch.zeros(
            (dm_view.shape[0], self.nfeats, ao.shape[-1]),
            device=dm.device,
            dtype=dm.dtype,
        )

        # Handle the density only case, where ao has one dim less
        if self.deriv == 0:
            c0 = dm_view @ ao
            features[..., 0, :] = torch.sum(c0 * ao[None, :, :], dim=-2)
            if len(dm.shape) == 2:
                return features.reshape((self.nfeats, -1))
            else:
                return features.reshape((*dm.shape[:-2], self.nfeats, -1))

        c0 = dm_view @ ao[0]

        feat_idx = 0
        if self.with_density:
            features[..., feat_idx, :] = torch.sum(c0 * ao[0][None, :, :], dim=-2)
            feat_idx += 1

        if self.with_grad:
            for i in range(3):
                features[..., feat_idx, :] = 2 * torch.sum(
                    c0 * ao[i + 1][None, :, :], dim=-2
                )
                feat_idx += 1

        if (self.with_kin or self.with_lapl) and not with_Q:
            for i in range(3):
                ci = dm_view @ ao[i + 1]
                features[..., feat_idx, :] += 0.5 * torch.sum(
                    ci * ao[i + 1][None, :, :], dim=-2
                )

            if self.with_kin:
                feat_idx += 1
                if self.with_lapl:
                    features[..., feat_idx, :] = 4 * features[..., feat_idx - 1, :]
            else:
                # Multiply times four for the laplacian
                features[..., feat_idx, :] *= 4.0

            if self.with_lapl:
                # 0 is without derivative
                # 1 2 3 are x y z derivatives
                # 4 5 6 are xx xy xz derivatives
                # 7 8 9 are yy yz zz derivatives
                for i in (4, 7, 9):
                    features[..., feat_idx, :] += 2 * torch.sum(
                        c0 * ao[i][None, :, :], dim=-2
                    )

        if with_Q:
            Q = torch.zeros(
                (dm_view.shape[0], ao.shape[-1], 3, 3), device=dm.device, dtype=dm.dtype
            )

            for i in range(3):
                ci = dm_view @ ao[i + 1]
                for j in range(i, 3):
                    Q = torch.sum(ci * ao[j + 1][None, :, :], dim=-2)

            if self.with_kin:
                features[..., feat_idx, :] = 0.5 * torch.einsum("...ii->...", Q)
                feat_idx += 1

            if self.with_lapl:
                features[..., feat_idx, :] = 2 * torch.einsum("...ii->...", Q)
                # 0 is without derivative
                # 1 2 3 are x y z derivatives
                # 4 5 6 are xx xy xz derivatives
                # 7 8 9 are yy yz zz derivatives
                for i in (4, 7, 9):
                    features[..., feat_idx, :] += 2 * torch.sum(
                        c0 * ao[i][None, :, :], dim=-2
                    )
                feat_idx += 1

            if self.with_ked_var:
                if not self.with_kin:
                    trace = torch.einsum("...ii->...", Q)
                else:
                    trace = 2 * features[:, feat_idx - 1, :]
                features[..., feat_idx, :] = 0.5 * torch.sum(
                    (
                        trace[:, None, None]
                        * torch.eye(3, device=dm.device, dtype=dm.dtype)[None, :, :]
                        - Q
                    )
                    ** 2,
                    dim=(-2, -1),
                )
                feat_idx += 1

            if self.with_ked_det:
                features[..., feat_idx, :] = torch.det(Q)
                feat_idx += 1
        if len(dm.shape) == 2:
            return features.reshape((self.nfeats, -1))
        else:
            return features.reshape((*dm.shape[:-2], self.nfeats, -1))


class ChunkEvalForward(Function):  # type: ignore[misc]
    @staticmethod
    def setup_context(
        ctx: FunctionCtx,
        inputs: tuple[
            torch.Tensor,
            gto.Mole,
            Grid,
            FeatureFunction,
            int,
            int,
            bool,
            bool,
            torch.Tensor,
        ],
        output: torch.Tensor,
    ) -> None:
        (
            ctx.dm,
            ctx.mol,
            ctx.grids,
            ctx.feature_function,
            ctx.blksize,
            ctx.compile_feature_function,
            ctx.gpu,
            *ctx.vectors_jvp,
        ) = inputs
        ctx.save_for_backward(ctx.dm)

    @staticmethod
    def forward(
        dm: torch.Tensor,
        mol: gto.Mole,
        grids: Grid,
        feature_function: FeatureFunction,
        blksize: int,
        compile_feature_function: bool,
        gpu: bool,
        *vectors_jvp: torch.Tensor,
    ) -> torch.Tensor:
        ngrids = grids.weights.size
        block_loop_args = (mol, grids, mol.nao)
        block_loop_kwargs = {
            "deriv": feature_function.deriv,
            "blksize": blksize if not gpu else None,
        }
        if gpu:
            check_gpu_imports_were_successful()
            ni = dft_gpu.numint.NumInt().build(mol, grids.coords)
            ni.grid_blksize = blksize
            sort_idx = ni.gdftopt._ao_idx
        else:
            ni = dft.numint.NumInt()
            sort_idx = np.arange(mol.nao_nr())

        features = torch.zeros(
            *dm.shape[:-2],
            feature_function.nfeats,
            ngrids,
            device=dm.device,
            dtype=dm.dtype,
        )
        if len(vectors_jvp) > 1 and feature_function.only_linear_feats:
            return features

        end = 0
        for ao_block, mask, weights, _ in ni.block_loop(
            *block_loop_args, **block_loop_kwargs
        ):
            start, end = end, end + weights.size
            # Mask dm to only include the relevant AOs
            if mask is None or not gpu:
                mask = torch.arange(mol.nao_nr(), device=dm.device)
            else:
                mask = torch.from_dlpack(mask)
            masked_dm = dm[..., sort_idx, :][..., sort_idx][
                ..., mask[:, None], mask[None, :]
            ]

            # Apply chain rule for this particular block
            partial_func = partial_feature_function_over_aos(
                feature_function,
                from_numpy_or_cupy(
                    ao_block, device=dm.device, dtype=dm.dtype, transpose=not gpu
                ),
            )
            for vector_jvp in vectors_jvp:
                partial_func = partial_jvp_function_over_tangents(
                    partial_func,
                    vector_jvp[..., sort_idx, :][..., sort_idx][
                        ..., mask[:, None], mask[None, :]
                    ],
                )

            # Compute feature (or its jvp) for this block with masked dm
            if compile_feature_function:
                temp_feature = torch.compile(partial_func)(masked_dm)
            else:
                temp_feature = partial_func(masked_dm)

            features[..., start:end] = temp_feature
        return features

    @staticmethod
    def jvp(ctx: FunctionCtx, grad_input: torch.Tensor) -> torch.Tensor:
        # Chain rule for the jvp
        return ChunkEvalForward.apply(
            ctx.dm,
            ctx.mol,
            ctx.grids,
            ctx.feature_function,
            ctx.blksize,
            ctx.compile_feature_function,
            ctx.gpu,
            *ctx.vectors_jvp,
            grad_input,
        )

    @staticmethod
    def backward(
        ctx: FunctionCtx, grad_output: torch.Tensor
    ) -> tuple[torch.Tensor | None, ...]:
        # After one vjp (backward) the signature of the function changes from dm.shape -> (*dm.shape[:-2], nfeats, ngrid) to dm.shape -> dm.shape
        # therefore we move to a different function that does essentially the same thing, but with the new signature

        # Derivative to dm
        grads = [
            ChunkEvalBackward.apply(
                ctx.dm,
                ctx.mol,
                ctx.grids,
                ctx.feature_function,
                ["jvp"] * len(ctx.vectors_jvp) + ["first_vjp"],
                ctx.blksize,
                ctx.compile_feature_function,
                ctx.gpu,
                *ctx.vectors_jvp,
                grad_output,
            )
        ]

        # We need to provide None for the gradients of the non-differentiable inputs
        # these are mol (1), grids (2), feature_function (3), blksize (4),
        # compile_feature_function (5), gpu (6)
        num_non_differentiable_inputs = 6

        grads += [None] * num_non_differentiable_inputs

        # Gradients of earlier tangents
        for i in range(len(ctx.vectors_jvp)):
            derivative_types = ["jvp"] * len(ctx.vectors_jvp)
            derivative_types[i] = "first_vjp"
            grads.append(
                ChunkEvalBackward.apply(
                    ctx.dm,
                    ctx.mol,
                    ctx.grids,
                    ctx.feature_function,
                    derivative_types,
                    ctx.blksize,
                    ctx.compile_feature_function,
                    ctx.gpu,
                    *ctx.vectors_jvp[:i],
                    grad_output,
                    *ctx.vectors_jvp[i + 1 :],
                )
            )

        return tuple(grads)


class ChunkEvalBackward(Function):  # type: ignore[misc]
    @staticmethod
    def setup_context(
        ctx: FunctionCtx,
        inputs: tuple[
            torch.Tensor,
            gto.Mole,
            Grid,
            FeatureFunction,
            list[str],
            int,
            bool,
            bool,
            torch.Tensor,
        ],
        output: tuple[torch.Tensor, ...],
    ) -> None:
        (
            ctx.dm,
            ctx.mol,
            ctx.grids,
            ctx.feature_function,
            ctx.derivative_types,
            ctx.blksize,
            ctx.compile_feature_function,
            ctx.gpu,
            *ctx.vectors,
        ) = inputs
        ctx.save_for_backward(ctx.dm)

    @staticmethod
    def forward(
        dm: torch.Tensor,
        mol: gto.Mole,
        grids: Grid,
        feature_function: FeatureFunction,
        derivative_types: list[str],
        blksize: int,
        compile_feature_function: bool,
        gpu: bool,
        *vectors: torch.Tensor,
    ) -> torch.Tensor:
        block_loop_args = (mol, grids, mol.nao)
        block_loop_kwargs = {
            "deriv": feature_function.deriv,
            "blksize": blksize if not gpu else None,
        }
        if gpu:
            check_gpu_imports_were_successful()
            ni = dft_gpu.numint.NumInt().build(mol, grids.coords)
            ni.grid_blksize = blksize
            sort_idx = ni.gdftopt._ao_idx
        else:
            ni = dft.numint.NumInt()
            sort_idx = np.arange(mol.nao_nr())

        end: int = 0
        out = torch.zeros_like(dm)
        if len(vectors) > 1 and feature_function.only_linear_feats:
            return out

        unsort_idx = torch.argsort(torch.tensor(sort_idx))
        for ao_block, mask, weights, _ in ni.block_loop(
            *block_loop_args,
            **block_loop_kwargs,
        ):
            start, end = end, end + weights.size

            # Mask to only include the relevant AOs
            if mask is None or not gpu:
                mask = torch.arange(mol.nao_nr(), device=dm.device)
            else:
                mask = from_numpy_or_cupy(mask, device=dm.device, dtype=torch.long)

            # Apply chain rule for this particular block
            # but be careful with signature change upon first vjp
            partial_func = partial_feature_function_over_aos(
                feature_function,
                from_numpy_or_cupy(
                    ao_block, device=dm.device, dtype=dm.dtype, transpose=not gpu
                ),
            )
            for derivative_type, vector in zip(derivative_types, vectors, strict=True):
                if derivative_type == "jvp":
                    partial_func = partial_jvp_function_over_tangents(
                        partial_func,
                        vector[..., sort_idx, :][..., sort_idx][
                            ..., mask[:, None], mask[None, :]
                        ],
                    )
                elif derivative_type == "vjp":
                    partial_func = partial_vjp_function_over_tangents(
                        partial_func,
                        vector[..., sort_idx, :][..., sort_idx][
                            ..., mask[:, None], mask[None, :]
                        ],
                    )
                elif derivative_type == "first_vjp":
                    partial_func = partial_vjp_function_over_tangents(
                        partial_func, vector[..., start:end]
                    )
                else:
                    raise ValueError(
                        f"Unknown derivative {derivative_type} (must be one of 'jvp', 'vjp', 'first_vjp')"
                    )
            if compile_feature_function:
                out[..., mask[:, None], mask[None, :]] += torch.compile(partial_func)(
                    dm[..., sort_idx, :][..., sort_idx][
                        ..., mask[:, None], mask[None, :]
                    ]
                )
            else:
                out[..., mask[:, None], mask[None, :]] += partial_func(
                    dm[..., sort_idx, :][..., sort_idx][
                        ..., mask[:, None], mask[None, :]
                    ]
                )
        return out[..., unsort_idx, :][..., unsort_idx]

    @staticmethod
    def jvp(ctx: FunctionCtx, *grad_input: torch.Tensor) -> torch.Tensor:
        # Chain rule for the jvp
        return ChunkEvalBackward.apply(
            ctx.dm,
            ctx.mol,
            ctx.grids,
            ctx.feature_function,
            ctx.derivative_types + ["jvp"],
            ctx.blksize,
            ctx.compile_feature_function,
            ctx.gpu,
            *ctx.vectors,
            grad_input,
        )

    @staticmethod
    def backward(
        ctx: FunctionCtx, grad_output: torch.Tensor
    ) -> tuple[torch.Tensor | None, ...]:
        # Chain rule for the vjp

        # Gradient corresponding to dm
        grads = [
            ChunkEvalBackward.apply(
                ctx.dm,
                ctx.mol,
                ctx.grids,
                ctx.feature_function,
                ctx.derivative_types + ["vjp"],
                ctx.blksize,
                ctx.compile_feature_function,
                ctx.gpu,
                *ctx.vectors,
                grad_output,
            )
        ]
        # We need to provide None for the gradients of the non-differentiable inputs
        # these are mol (1), grids (2), feature_function (3), derivative_types (4), blksize (5),
        # compile_feature_function (6), gpu (7)
        num_non_differentiable_inputs = 7

        grads += [None] * num_non_differentiable_inputs
        # Gradients of gradients
        for i, derivative_type in enumerate(ctx.derivative_types):
            derivative_types = copy(ctx.derivative_types)
            if derivative_type == "jvp" or derivative_type == "vjp":
                derivative_types[i] = "vjp"
                grads.append(
                    ChunkEvalBackward.apply(
                        ctx.dm,
                        ctx.mol,
                        ctx.grids,
                        ctx.feature_function,
                        derivative_types,
                        ctx.blksize,
                        ctx.compile_feature_function,
                        ctx.gpu,
                        *ctx.vectors[:i],
                        grad_output,
                        *ctx.vectors[i + 1 :],
                    )
                )
            elif derivative_type == "first_vjp":
                grads.append(
                    ChunkEvalForward.apply(
                        ctx.dm,
                        ctx.mol,
                        ctx.grids,
                        ctx.feature_function,
                        ctx.blksize,
                        ctx.compile_feature_function,
                        ctx.gpu,
                        *ctx.vectors[:i],
                        grad_output,
                        *ctx.vectors[i + 1 :],
                    )
                )
            else:
                raise ValueError(
                    f"Unknown derivative {derivative_type} (must be one of 'jvp', 'vjp', 'first_vjp')"
                )
        return tuple(grads)


def non_chunk(
    dm: torch.Tensor,
    mol: gto.Mole,
    grids: Grid,
    feature_function: FeatureFunction,
    compile_feature_function: bool = False,
    gpu: bool = False,
) -> torch.Tensor:
    if gpu:
        check_gpu_imports_were_successful()
        ni = dft_gpu.numint.NumInt().build(mol, grids.coords)
    else:
        ni = dft.numint.NumInt()
    ao = from_numpy_or_cupy(
        ni.eval_ao(mol, grids.coords, deriv=feature_function.deriv, non0tab=None),
        device=dm.device,
        dtype=dm.dtype,
        transpose=True,
    )
    if compile_feature_function:
        return torch.compile(feature_function.forward)(dm, ao)
    else:
        return feature_function.forward(dm, ao)


def auto_chunk(
    dm: torch.Tensor,
    mol: gto.Mole,
    grids: Grid,
    feature_function: FeatureFunction,
    block_size: int | None = None,
    max_memory: int = 2000,
    fix_block_size: bool = True,
    compile_feature_function: bool = False,
    gpu: bool = False,
) -> dict[str, torch.Tensor]:
    """
    Automatically splits feature evaluation into smaller chunks if needed.

    This function determines the appropriate chunk size for evaluating a feature
    function on molecular grids, based on available memory and number of basis
    functions. If the computed chunk size is larger than the size of the grid, or
    if a fixed block size was provided, it uses a non-chunked approach.

    Parameters
    ----------
    dm: torch.Tensor
        Density matrix or set of density matrices used for
        evaluating the feature function.
    mol: gto.Mole
        PySCF molecule object representing the system of interest.
    grids: Grid
        Grids object defining the points in space on which
        the feature function is evaluated.
    feature_function: FeatureFunction
        The object representing the feature function to evaluate. The number of derivatives (deriv) determines
        how many components to compute.
    gpu: bool, optional
        Whether to use GPU for computation. Defaults to False.
    block_size: int | None, optional
        Manually specified block size for chunking. (CPU only)
        Defaults to None.
    max_memory: int, optional
        Maximum memory in MB to use for chunking (CPU only)
    fix_block_size: bool, optional
        Whether to fix the block size or compute it
        automatically based on system resources. Defaults to True. (CPU only)
    compile_feature_function: bool, optional
        If True, compiles the feature function for efficiency. Defaults to False.

    Returns
    -------
    dict[str, torch.Tensor]:
        The evaluated feature function on the specified grids, either
        computed in smaller chunks or in a single pass, depending on the block size.
    """

    if gpu:
        check_gpu_imports_were_successful()
        if dm.device.type != "cuda":
            raise ValueError("Density matrix must be on the GPU when gpu=True.")

    blksize: int | None

    if gpu and block_size is not None:
        raise ValueError("Setting custom block size is not supported on GPU.")

    if block_size is None and fix_block_size and not gpu:
        nao = mol.nao_nr()
        comp = (
            (feature_function.deriv + 1)
            * (feature_function.deriv + 2)
            * (feature_function.deriv + 3)
            // 6
        )
        BLKSIZE = dft.gen_grid.BLKSIZE
        blksize = int(max_memory * 1e6 / ((comp + 1) * nao * 8 * BLKSIZE))
        blksize = max(4, min(blksize, 1200)) * BLKSIZE
    else:
        blksize = block_size

    if blksize is not None and not gpu:
        blksize = blksize - blksize % dft.gen_grid.BLKSIZE

    if blksize is not None and blksize >= grids.weights.shape[0]:
        features = non_chunk(
            dm.double(),
            mol,
            grids,
            feature_function,
            compile_feature_function=compile_feature_function,
            gpu=gpu,
        )
    else:
        features = ChunkEvalForward.apply(
            dm.double(),
            mol,
            grids,
            feature_function,
            blksize,
            compile_feature_function,
            gpu,
        )
    return feature_function.to_dict(features)



# SPDX-License-Identifier: MIT
# Copyright (c) Microsoft Corporation.
# Copied from src/skala/pyscf/backend.py

from typing import (
    TYPE_CHECKING,
    TypeAlias,
    TypeVar,
)

import numpy as np
import torch
from pyscf import dft
from torch import Tensor

GPU_EXCEPTION: BaseException | None = None

__all__ = [
    "Array",
    "Grid",
    "KS",
    "dft_gpu",
    "check_gpu_imports_were_successful",
    "from_numpy_or_cupy",
    "to_numpy",
    "to_cupy",
]


if TYPE_CHECKING:
    # During type checking, we do the same as during normal runtime, but without the try/except.
    import cupy
    from gpu4pyscf import dft as dft_gpu

    GPU_EXCEPTION = None

    Array = TypeVar("Array", np.ndarray, cupy.ndarray)
    Grid: TypeAlias = dft.Grids | dft_gpu.Grids
    KS: TypeAlias = dft.rks.RKS | dft.uks.UKS | dft_gpu.rks.RKS | dft_gpu.uks.UKS

    # Reset the default CuPy memory allocator to avoid memory leak issues
    # that seem to arise when combining the custom allocator of gpu4pyscf with DLPack usage.
    cupy.cuda.set_allocator(cupy.get_default_memory_pool().malloc)
else:
    try:
        import cupy
        from gpu4pyscf import dft as dft_gpu

        Array = TypeVar("Array", np.ndarray, cupy.ndarray)
        Grid: TypeAlias = dft.Grids | dft_gpu.Grids
        KS: TypeAlias = dft.rks.RKS | dft.uks.UKS | dft_gpu.rks.RKS | dft_gpu.uks.UKS

        # Reset the default CuPy memory allocator to avoid memory leak issues
        # that seem to arise when combining the custom allocator of gpu4pyscf with DLPack usage.
        cupy.cuda.set_allocator(cupy.get_default_memory_pool().malloc)
        GPU_EXCEPTION = None
    except ImportError as e:
        GPU_EXCEPTION = e
        dft_gpu = None

        Array = TypeVar("Array", bound=np.ndarray)
        Grid: TypeAlias = dft.Grids
        KS: TypeAlias = dft.rks.RKS | dft.uks.UKS


def check_gpu_imports_were_successful() -> None:
    if GPU_EXCEPTION is not None:
        raise GPU_EXCEPTION


def from_numpy_or_cupy(
    x: Array,
    *,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
    transpose: bool = False,
) -> Tensor:
    if isinstance(x, np.ndarray):
        x_torch = torch.from_numpy(x)
    else:
        x_torch = torch.from_dlpack(x)
    x_torch = x_torch.to(device=device, dtype=dtype)
    if transpose:
        return x_torch.transpose(-1, -2)
    else:
        return x_torch


def to_numpy(x: Tensor) -> np.ndarray:
    return x.detach().cpu().numpy()


def to_cupy(x: Tensor) -> "cupy.ndarray":
    return cupy.from_dlpack(x.detach())
